# temperature conversions
zero_celcius = 273.15     # 0 degrees C in K

# reference values
t_ref = 293.15            # reference temperature 20 degrees celcius
p_ref = 1013.25           # reference pressure in mbar
z_ref = 0                 # sea level m

lapse = -0.0065           # lapse rate K m-1
g = 9.807                 # gravity m s-2
gc_spec = 287.0           # gas constant J kg-1 K-1
gc_dry = 2.87             # dry air gas constant mbar K-1 m3 kg-1
gc_moist = 4.61           # moist air gas constant mbar K-1 m3 kg-1
r_mw = 0.622              # ratio water particles/ air particles
sh = 1004.0               # specific heat J kg-1 K-1
lh_0 = 2501000.0          # latent heat of evaporation at 0 C [J/kg]
lh_rate = -2361           # rate of latent heat vs temperature [J/kg/C]
power = (g/(-lapse*gc_spec))
k = 0.41                  # karman constant (-)
sol = 1367                # maximum solar radiation at top of atmosphere W m-2
sb = 5.67e-8              # stefan boltzmann constant
day_sec = 86400.0         # seconds in a day
year_sec = day_sec * 365  # seconds in a year

absorbed_radiation = 0.48 # biomass factor
conversion = 0.864        # conversion biomass calculation from g s-1 m-2
                          # to kg ha-1 d-1

z0_soil = 0.001           # soil roughness m

#############
#############
#############

aod550_i = 0.01 # https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/61/MOD04_L2 heb niet echt een standaard product hiervan gevonden
se_top = 0.5
porosity = 0.4

# http://lawr.ucdavis.edu/classes/SSC100/probsets/pset01.html
# 6. Calculate the porosity of a soil sample that has a bulk density of 1.35 g/cm3. Assume the particle density is 2.65 g/cm3.
# Porosity = (1-(r b/r d) x 100 = (1-(1.35/2.65)) x 100 = 49%
# page 31 flow diagram

# **effective_leaf_area_index**************************************************
# constants or predefined:
nd_min = 0.1 # 0.125
nd_max = 0.8
vc_pow = 0.7
vc_min = 0
vc_max = 0.9677324224821418
lai_pow = -0.45

# **atmospheric canopy resistance***********************************************
# constants or predefined:
diffusion_slope = -1.33
diffusion_intercept = 1.15
t_opt = 25 # optimal temperature for plant growth
t_min = 0 # minimal temperature for plant growth
t_max = 50 # maximal temperature for plant growth
vpd_slope = -0.3
rcan_max = 1000000

# **net radiation canopy******************************************************
# constants or predefined:
vp_slope = 0.14
vp_offset = 0.34
lw_slope = 1.35
lw_offset = 0.35
int_max = 0.2

# **canopy resistance***********************************************************
# constants or predefined:
r0_bare = 0.42 # 0.38
r0_bare_wet = 0.20
r0_full = 0.18
tenacity = 1.0
disp_bare = 0.0
disp_full = 0.667
fraction_h_bare = 0.65
fraction_h_full = 0.95

z_obs = 2
z_b = 100
z0m_bare = 0.0005 # 0.001

# **initial canopy aerodynamic resistance***********************************************************
# constants or predefined:
ndvi_obs_min = 0.25
ndvi_obs_max = 0.75
obs_fr = 0.25

# **ETLook.unstable.initial_friction_velocity_daily***********************************************************
# constants or predefined:
c1 = 1

# **ETLook.unstable.transpiration***********************************************************
# constants or predefined:
iter_h = 3

# **ETLook.resistance.soil_resistance***********************************************************
# constants or predefined:
r_soil_pow = -2.1
r_soil_min = 800

# **ETLook.unstable.initial_sensible_heat_flux_soil_daily***********************************************************
# constants or predefined:
#porosity = 0.4 #Note: soil dependent
#se_top = 1.0 #Note should be input !
rn_slope = 0.92
rn_offset = -61.0

# **ETLook.unstable.evaporation***********************************************************
# constants or predefined:
r0_grass = 0.23


import os
from dask.diagnostics import ProgressBar
import numpy as np
from pywapor.general.logger import log
import xarray as xr
import numpy as np
import shutil
import glob
from datetime import datetime as dt
import warnings
import rasterio.warp
import pandas as pd
from pywapor.general.performance import performance_check

def log_example_ds(example_ds):
    """Writes some metadata about a `example_ds` to the logger.

    Parameters
    ----------
    example_ds : xr.Dataset
        Dataset for which to log information.
    """
    if "source" in example_ds.encoding.keys():
        log.info(f"--> Using `{os.path.split(example_ds.encoding['source'])[-1]}` as reprojecting example.").add()
    else:
        log.info(f"--> Using variable `{list(example_ds.data_vars)[0]}` as reprojecting example.").add()
    shape = example_ds.y.size, example_ds.x.size
    res = example_ds.rio.resolution()
    log.info(f"> shape: {shape}, res: {abs(res[0]):.4f}° x {abs(res[1]):.4f}°.").sub()

def adjust_timelim_dtype(timelim):
    """Convert different time formats to `datetime.datetime`.

    Parameters
    ----------
    timelim : list
        List defining a period.

    Returns
    -------
    list
        List defining a period using `datetime.datetime` objects.
    """
    if isinstance(timelim[0], str):
        timelim[0] = dt.strptime(timelim[0], "%Y-%m-%d")
        timelim[1] = dt.strptime(timelim[1], "%Y-%m-%d")
    elif isinstance(timelim[0], np.datetime64):
        timelim[0] = dt.utcfromtimestamp(timelim[0].tolist()/1e9).date()
        timelim[1] = dt.utcfromtimestamp(timelim[1].tolist()/1e9).date()
    return timelim

def remove_ds(ds):
    """Delete a dataset-file from disk, assuring it's closed properly before doing so.

    Parameters
    ----------
    ds : xr.Dataset | str
        Either a `xr.Dataset` (in which case its `source` as defined in the `encoding` attribute will be used)
        or a `str` in which case it must be a path to a file.
    """
    if isinstance(ds, xr.Dataset):
        if "source" in ds.encoding.keys():
            fp = ds.encoding["source"]
            ds = xr.open_dataset(fp)
            ds = ds.close()
            os.remove(fp)
    elif isinstance(ds, str):
        if os.path.isfile(ds):
            fp = ds
            ds = xr.open_dataset(fp)
            ds = ds.close()
            os.remove(fp)        

def process_ds(ds, coords, variables, crs = None):
    """Apply some rioxarray related transformations to a dataset.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to be processed.
    coords : dict
        Dictionary describing the names of the spatial coordinates.
    variables : list
        List of variables to keep.
    crs : rasterio.CRS.crs, optional
        Coordinate reference system to assign (no reprojection is done), by default None.

    Returns
    -------
    xr.Dataset
        Dataset with some attributes corrected.
    """

    ds = ds[list(variables.keys())]

    ds = ds.rename({v[0]:k for k,v in coords.items() if k in ["x", "y"]})
    ds = ds.rename({k: v[1] for k, v in variables.items()})

    if not isinstance(crs, type(None)):
        ds = ds.rio.write_crs(crs)

    ds = ds.rio.write_grid_mapping("spatial_ref")

    for var in [x for x in list(ds.variables) if x not in ds.coords]:
        if "grid_mapping" in ds[var].attrs.keys():
            del ds[var].attrs["grid_mapping"]

    ds = ds.sortby("y", ascending = False)
    ds = ds.sortby("x")

    ds.attrs = {}

    return ds

def make_example_ds(ds, folder, target_crs, bb = None):
    """Make an dataset suitable to use as an example for matching with other datasets.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset.
    folder : str
        Path to folder in which to save the `example_ds`.
    target_crs : rasterio.CRS.crs
        Coordinate reference system of the `example_ds`.
    bb : list, optional
        Boundingbox of the `example_ds` ([xmin, ymin, xmax, ymax]), by default None.

    Returns
    -------
    xr.Dataset
        Example dataset.
    """
    example_ds_fp = os.path.join(folder, "example_ds.nc")
    if os.path.isfile(example_ds_fp):
        example_ds = open_ds(example_ds_fp)
    else:
        if not isinstance(bb, type(None)):
            if ds.rio.crs != target_crs:
                loc_bb = transform_bb(target_crs, ds.rio.crs, bb)
            else:
                loc_bb = bb
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category = FutureWarning)
                ds = ds.rio.clip_box(*loc_bb)
            ds = ds.rio.pad_box(*loc_bb)
        ds = ds.rio.reproject(target_crs)
        if not isinstance(bb, type(None)):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category = UserWarning)
                ds = ds.rio.clip_box(*bb)
        ds = ds.drop_vars(list(ds.data_vars))
        example_ds = save_ds(ds, example_ds_fp, encoding = "initiate", label = f"Creating example dataset.") # NOTE saving because otherwise rio.reproject bugs.
    return example_ds

@performance_check
def save_ds(ds, fp, decode_coords = "all", encoding = None, chunks = "auto", precision = 4):
    """Save a `xr.Dataset` as netcdf.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to save.
    fp : str
        Path to file to create.
    decode_coords : str, optional
        Controls which variables are set as coordinate variables when
        reopening the dataset, by default `"all"`.
    encoding : "initiate" | dict | None, optional
        Apply an encoding to the saved dataset. "initiate" will create a encoding on-the-fly, by default None.
    chunks : "auto" | dict
        Define the chunks with which to perform any pending calculations, by default "auto".
    precision : int | dict, optional
        How many decimals to store for each variable, only used when `encoding` is `"initiate"`, by default 4.

    Returns
    -------
    xr.Dataset
        The newly created dataset.
    """
    temp_fp = fp.replace(".nc", "_temp.xx")

    folder = os.path.split(fp)[0]
    if not os.path.isdir(folder):
        os.makedirs(folder)

    valid_coords = ["x", "y", "spatial_ref", "time", "time_bins"]
    for coord in ds.coords.values():
        if coord.name not in valid_coords:
            ds = ds.drop_vars([coord.name])

    if isinstance(chunks, dict):
        chunks = {dim: v for dim, v in chunks.items() if dim in ds.dims}

    ds = ds.chunk(chunks)

    if "y" in ds.coords:
        if len(ds.y.dims) == 1:
            ds = ds.sortby("y", ascending = False)
        ds = ds.rio.write_transform(ds.rio.transform(recalc=True))

    if encoding == "initiate":
        if not isinstance(precision, dict):
            precision = {var: precision for var in ds.data_vars}
        encoding = {var: {
                        "zlib": True,
                        "_FillValue": -9999,
                        "chunksizes": tuple([v[0] for _, v in ds[var].chunksizes.items()]),
                        "dtype": "int32", # determine_dtype(ds[var], -9999, precision.get(var)),
                        "scale_factor": 10**-precision.get(var, 0), 
                        } for var in ds.data_vars if np.all([spat in ds[var].coords for spat in ["x", "y"]])}
        for var in ds.data_vars:
            if "_FillValue" in ds[var].attrs.keys():
                _ = ds[var].attrs.pop("_FillValue")
        if "spatial_ref" in ds.coords:
            for var in ds.data_vars:
                if np.all([spat in ds[var].coords for spat in ["x", "y"]]):
                    ds[var].attrs.update({"grid_mapping": "spatial_ref"})
        for var in ds.coords:
            if var in ds.dims:
                encoding[var] = {"dtype": "float64"}

    with ProgressBar(minimum = 90*10, dt = 2.0):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            ds.to_netcdf(temp_fp, encoding = encoding)

    ds = ds.close()

    os.rename(temp_fp, fp)

    ds = open_ds(fp, decode_coords = decode_coords, chunks = chunks)

    return ds

def open_ds(fp, decode_coords = "all", chunks = "auto", **kwargs):
    """Open a file using xarray.

    Parameters
    ----------
    fp : str
        Path to file.
    decode_coords : str, optional
        Whether or not to decode coordinates, by default "all".
    chunks : str | dict, optional
        Chunks to use, by default "auto".

    Returns
    -------
    xr.Dataset
        Opened file.
    """
    ds = xr.open_dataset(fp, decode_coords = decode_coords, chunks = chunks, **kwargs)
    return ds

def create_dummy_mask(x, y, sign = None, slope = None, xshift_fact = None, yshift_fact = None):
    if isinstance(sign, type(None)):
        sign = np.sign(np.random.random() - 0.5)
    if isinstance(slope, type(None)):
        slope = np.random.random()
    if isinstance(xshift_fact, type(None)):
        xshift_fact = np.random.random()
    if isinstance(yshift_fact, type(None)):
        yshift_fact = np.random.random()
    slope = sign*(slope * np.ptp(y) / np.ptp(x))
    yshift = {-1: y.max(), 1: y.min()}[np.sign(slope)]
    xshift = x.min()
    mask = (x - (xshift + xshift_fact * np.ptp(x))) * slope + (yshift + sign * yshift_fact * np.ptp(y))
    return y < mask

def create_dummy_ds(varis, fp = None, shape = (10, 1000, 1000), chunks = (-1, 500, 500), 
                    sdate = "2022-02-01", edate = "2022-02-11", precision = 2, min_max = [-1, 1],
                    latlim = [20,30], lonlim = [40, 50], data_generator = "random", mask_data = False):
    check = False
    if not isinstance(fp, type(None)):
        if os.path.isfile(fp):
            os.remove(fp)
    if not check:
        nt, ny, nx = shape
        dates = pd.date_range(sdate, edate, periods = nt)
        y,t,x = np.meshgrid(np.linspace(latlim[0], latlim[1], shape[1]), np.linspace(0, len(dates) + 1, len(dates)), np.linspace(lonlim[0],lonlim[1],shape[2]))
        if data_generator == "random":
            data = np.random.uniform(size = np.prod(shape), low = min_max[0], high=min_max[1]).reshape(shape)
        elif data_generator == "uniform":
            data = np.sqrt(x**2 + y**2)
            data = (data - data.min()) * ((min_max[1] - min_max[0]) / (data.max() - data.min())) + min_max[0]
        if mask_data:
            for i in range(nt):
                mask = create_dummy_mask(x[0,...], y[0,...])
                data[i,mask] = np.nan
        ds = xr.Dataset({k: (["time", "y", "x"], data) for k in varis}, coords = {"time": dates, "y": np.linspace(latlim[0], latlim[1], ny), "x": np.linspace(lonlim[0], lonlim[1], nx)})
        ds = ds.rio.write_crs(4326)
        if isinstance(chunks, tuple):
            chunks = {name: size for name, size in zip(["time", "y", "x"], chunks)}
        if not isinstance(fp, type(None)):
            ds = save_ds(ds, fp, chunks = chunks, encoding = "initiate", precision = precision, label = "Creating dummy dataset.")
    return ds

def determine_dtype(da, nodata, precision = None):
    if isinstance(precision, type(None)) and da.dtype.kind == "f":
        dtypes = [np.float16, np.float32, np.float64]
        precision = 0
        info = np.finfo
    else:
        dtypes = [np.int8, np.int16, np.int32, np.int64]
        info = np.iinfo
        if isinstance(precision, type(None)):
            precision = 0
    low = np.min([nodata, np.int0(np.floor(da.min().values * 10**precision))]) > [info(x).min for x in dtypes]
    high = np.max([nodata, np.int0(np.ceil(da.max().values * 10**precision))]) < [info(x).max for x in dtypes]
    check = np.all([low, high], axis = 0)
    if True in check:
        dtype = dtypes[np.argmax(check)]
    else:
        dtype = dtypes[-1]
        log.warning(f"--> Data for `{da.name}` with range [{np.int0(np.floor(da.min().values))}, {np.int0(np.ceil(da.max().values))}] doesnt fit inside dtype {dtype} with range [{np.iinfo(dtype).min}, {np.iinfo(dtype).max}] with a {precision} decimal precision.")
    return np.dtype(dtype).name

def create_wkt(latlim, lonlim):
    left = lonlim[0]
    bottom = latlim[0]
    right = lonlim[1]
    top = latlim[1]
    x = f"{left} {bottom},{right} {bottom},{right} {top},{right} {bottom},{left} {bottom}"
    return "GEOMETRYCOLLECTION(POLYGON((" + x + ")))"

def unpack(file, folder):
    fn = os.path.splitext(file)[0]
    shutil.unpack_archive(os.path.join(folder, file), folder)
    folder = [x for x in glob.glob(os.path.join(folder, fn + "*")) if os.path.isdir(x)][0]
    return folder

def transform_bb(src_crs, dst_crs, bb):
    """Transforms coordinates from one CRS to another.

    Parameters
    ----------
    src_crs : rasterio.CRS.crs
        Source CRS.
    dst_crs : rasterio.CRS.crs
        Target CRS.
    bb : list
        Coordinates to be transformed.

    Returns
    -------
    list
        Transformed coordinates.
    """
    bb =rasterio.warp.transform_bounds(src_crs, dst_crs, *bb, densify_pts=21)
    return bb

def calc_dlat_dlon(geo_out, size_X, size_Y, lat_lon = None):
    """
    Calculated the dimensions of each pixel in meter.

    Parameters
    ----------
    geo_out: list
        Geotransform function of the array.
    size_X: int
        Number of pixels in x-direction.
    size_Y: int
        Number of pixels in y-direction.
    lat_lon : tuple, optional
        Tuple with two rasters, one for latitudes, another for longitudes, by default None.

    Returns
    -------
    tuple
        Tuple with two arrays with teh size of each pixel in the x and y direction in meters.
    """
    if isinstance(lat_lon, type(None)):
        # Create the lat/lon rasters
        lon = np.arange(size_X + 1)*geo_out[1]+geo_out[0] - 0.5 * geo_out[1]
        lat = np.arange(size_Y + 1)*geo_out[5]+geo_out[3] - 0.5 * geo_out[5]
    else:
        lat, lon = lat_lon

    dlat_2d = np.array([lat,]*int(np.size(lon,0))).transpose()
    dlon_2d =  np.array([lon,]*int(np.size(lat,0)))

    # Radius of the earth in meters
    R_earth = 6371000

    # Calculate the lat and lon in radians
    lonRad = dlon_2d * np.pi/180
    latRad = dlat_2d * np.pi/180

    # Calculate the difference in lat and lon
    lonRad_dif = abs(lonRad[:,1:] - lonRad[:,:-1])
    latRad_dif = abs(latRad[:-1] - latRad[1:])

    # Calculate the distance between the upper and lower pixel edge
    a = np.sin(latRad_dif[:,:-1]/2) * np.sin(latRad_dif[:,:-1]/2)
    clat = 2 * np.arctan2(np.sqrt(a), np.sqrt(1-a))
    dlat = R_earth * clat

    # Calculate the distance between the eastern and western pixel edge
    b = np.cos(latRad[1:,:-1]) * np.cos(latRad[:-1,:-1]) * np.sin(lonRad_dif[:-1,:]/2) * np.sin(lonRad_dif[:-1,:]/2)
    clon = 2 * np.arctan2(np.sqrt(b), np.sqrt(1-b))
    dlon = R_earth * clon

    return(dlat, dlon)

if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Local/dummy_ds_test"

    varis = ["my_var"]
    shape = (10, 1000, 1000)
    sdate = "2022-02-02"
    edate = "2022-02-13" 
    fp = os.path.join(folder, "dummy_test.nc")
    precision = 2
    min_max = [-1, 1]

    ds = create_dummy_ds(varis, 
                    shape = shape, 
                    sdate = sdate, 
                    edate = edate, 
                    fp = fp,
                    precision = precision,
                    min_max = min_max,
                    )


import numpy as np
import pywapor.general.processing_functions as pf
import xarray as xr
from pywapor.general.logger import log

def calc_slope(ds, var):
    """Calculate the gradient of `z`.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing var.
    var : str
        Variable name.

    Returns
    -------
    xr.Dataset
        Enhanced dataset.
    """
    reqs = ["z", "y", "x"]
    if np.all([x in ds.variables for x in reqs]):

        dlat, dlon = pf.calc_dlat_dlon(None, None, None, (ds.y.values, ds.x.values))            
        new_x_coords = np.append([0], np.cumsum(np.repeat(np.nanmean(dlon), ds["z"].sizes["x"]-1)))
        new_y_coords = np.append([0], np.cumsum(np.repeat(np.nanmean(dlat), ds["z"].sizes["y"]-1)))
        temp_ds = ds.drop_vars(["x", "y"]).assign_coords({"x": new_x_coords, "y": new_y_coords})

        y = temp_ds["z"].differentiate("x")
        x = temp_ds["z"].differentiate("y")

        y = xr.where((y >= -1e-10) & (y <= 1e-10), 0, y)
        x = xr.where((x >= -1e-10) & (x <= 1e-10), 0, x)

        hypot = np.hypot(x,y)
        slope = np.arctan(hypot) * 180.0 / np.pi

        ds[var] = slope.drop_vars(["x", "y"])
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")

    return ds

def calc_aspect(ds, var):
    """Calculate the aspect of `z`.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing var.
    var : str
        Variable name.

    Returns
    -------
    xr.Dataset
        Enhanced dataset.
    """
    reqs = ["z", "y", "x"]
    if np.all([x in ds.variables for x in reqs]):

        dlat, dlon = pf.calc_dlat_dlon(None, None, None, (ds.y.values, ds.x.values))            
        new_x_coords = np.append([0], np.cumsum(np.repeat(np.nanmean(dlon), ds["z"].sizes["x"]-1)))
        new_y_coords = np.append([0], np.cumsum(np.repeat(np.nanmean(dlat), ds["z"].sizes["y"]-1)))
        temp_ds = ds.drop_vars(["x", "y"]).assign_coords({"x": new_x_coords, "y": new_y_coords})

        y = temp_ds["z"].differentiate("x")
        x = temp_ds["z"].differentiate("y")

        y = xr.where((y >= -1e-10) & (y <= 1e-10), 0, y)
        x = xr.where((x >= -1e-10) & (x <= 1e-10), 0, x)

        aspect = np.arctan2(y/np.nanmean(dlat), -x/np.nanmean(dlon))  * 180.0 / np.pi + 180.0

        ds[var] = aspect.drop_vars(["x", "y"])
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")

    return ds



import glob
import os
import xarray as xr
import numpy as np
from datetime import datetime as dt
from pywapor.general.logger import log, adjust_logger
import pywapor.collect.protocol.sentinelapi as sentinelapi
import numpy as np
from functools import partial
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds

def apply_qa(ds, var):
    """Mask SENTINEL2 data using a qa variable.

    Parameters
    ----------
    ds : xr.Dataset
        Input data, should contain `"qa"` and `var` as variables.
    var : str
        Variable name in `ds` to be masked.

    Returns
    -------
    xr.Dataset
        Masked dataset.
    """
    # 0 SC_NODATA # 1 SC_SATURATED_DEFECTIVE # 2 SC_DARK_FEATURE_SHADOW
    # 3 SC_CLOUD_SHADOW # 4 SC_VEGETATION # 5 SC_NOT_VEGETATED
    # 6 SC_WATER # 7 SC_UNCLASSIFIED # 8 SC_CLOUD_MEDIUM_PROBA
    # 9 SC_CLOUD_HIGH_PROBA # 10 SC_THIN_CIRRUS # 11 SC_SNOW_ICE
    if "qa" in ds.data_vars:
        pixel_qa_flags = [0, 1, 2, 3, 7, 8, 9, 10, 11]
        keep = np.invert(ds["qa"].isin(pixel_qa_flags))
        ds[var] = ds[var].where(keep)
    else:
        log.warning(f"--> Couldn't apply qa, since `qa` doesn't exist in this dataset ({list(ds.data_vars)}).")
    return ds

def mask_invalid(ds, var, valid_range = (1, 65534)):
    """Mask invalid data.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Variable to mask.
    valid_range : tuple, optional
        Range of valid values in `var`, by default (1, 65534).

    Returns
    -------
    xr.Dataset
        Masked data.
    """
    # 0 = NODATA, 65535 = SATURATED
    ds[var] = ds[var].where((ds[var] >= valid_range[0]) & (ds[var] <= valid_range[1]))
    return ds

def scale_data(ds, var):
    """Apply a scale and offset factor to `ds`.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset.
    var : str
        Variable to scale and offset.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    scale = 1./10000. # BOA_QUANTIFICATION_VALUE
    offset = -1000 # BOA_ADD_OFFSET
    ds[var] = (ds[var] + offset) * scale
    ds[var] = ds[var].where((ds[var] <= 1.00) & (ds[var] >= 0.00))
    return ds

def calc_normalized_difference(ds, var, bands = ["nir", "red"]):
    """Calculate the normalized difference of two bands.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.
    bands : list, optional
        The two bands to use to calculate the norm. difference, by default ["nir", "red"].

    Returns
    -------
    xr.Dataset
        Output data.
    """
    if np.all([x in ds.data_vars for x in bands]):
        da = (ds[bands[0]] - ds[bands[1]]) / (ds[bands[0]] + ds[bands[1]])
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in bands if x not in ds.data_vars])}` is missing.")
    return ds

def calc_psri(ds, var):
    """Calculate the PSRI as ("red" - "blue)/"red_edge_740".

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["red", "blue", "red_edge_740"]
    if np.all([x in ds.data_vars for x in reqs]):
        da = (ds["red"] - ds["blue"]) / ds["red_edge_740"]
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def calc_nmdi(ds, var):
    """Calculate the NMDI.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["swir1", "swir2", "nir"]
    if np.all([x in ds.data_vars for x in reqs]):
        ds["nominator"] = ds["swir1"] - ds["swir2"]
        ds = calc_normalized_difference(ds, var, bands = ["nominator", "nir"])
        ds = ds.drop_vars(["nominator"])
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def calc_bsi(ds, var):
    """Calculate the BSI.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["nir", "swir1", "red", "blue"]
    if np.all([x in ds.data_vars for x in reqs]):
        ds["nominator"] = ds["nir"] + ds["blue"]
        ds["denominator"] = ds["swir1"] + ds["red"]
        ds = calc_normalized_difference(ds, var, bands = ["nominator", "denominator"])
        ds = ds.drop_vars(["nominator", "denominator"])
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def calc_vari_red_egde(ds, var):
    """Calculate the VARI_RED_EDGE.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    reqs = ["red_edge_740", "blue", "red"]
    if np.all([x in ds.data_vars for x in reqs]):
        n1 = ds["red_edge_740"] - 1.7 * ds["red"] + 0.7 * ds["blue"]
        n2 = ds["red_edge_740"] + 2.3 * ds["red"] - 1.3 * ds["blue"]
        da = n1 / n2
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def calc_r0(ds, var):
    """Calculate the Albedo.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """
    weights = {
        "blue": 0.074,
        "green": 0.083,
        "red": 0.334,
        "nir": 0.356,
        "offset": 0.033,
    }
    reqs = ["blue", "green", "red", "nir"]
    if np.all([x in ds.data_vars for x in reqs]):
        ds["offset"] = xr.ones_like(ds["blue"])
        weights_da = xr.DataArray(data = list(weights.values()), 
                                coords = {"band": list(weights.keys())})
        ds["r0"] = ds[reqs + ["offset"]].to_array("band").weighted(weights_da).sum("band", skipna = False)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {

        "S2MSI2A_R20m": {
                    "_B01_20m.jp2": [(), "coastal_aerosol", [mask_invalid, apply_qa, scale_data]],
                    "_B02_20m.jp2": [(), "blue", [mask_invalid, apply_qa, scale_data]],
                    "_B03_20m.jp2": [(), "green", [mask_invalid, apply_qa, scale_data]],
                    "_B04_20m.jp2": [(), "red", [mask_invalid, apply_qa, scale_data]],
                    "_B05_20m.jp2": [(), "red_edge_703", [mask_invalid, apply_qa, scale_data]],
                    "_B06_20m.jp2": [(), "red_edge_740", [mask_invalid, apply_qa, scale_data]],
                    "_B07_20m.jp2": [(), "red_edge_782", [mask_invalid, apply_qa, scale_data]],
                    "_B8A_20m.jp2": [(), "nir", [mask_invalid, apply_qa, scale_data]],
                    "_B11_20m.jp2": [(), "swir1", [mask_invalid, apply_qa, scale_data]],
                    "_B12_20m.jp2": [(), "swir2", [mask_invalid, apply_qa, scale_data]],
                    "_SCL_20m.jp2": [(), "qa", []],
                },
        "S2MSI2A_R60m": {
                    "_B01_60m.jp2": [(), "coastal_aerosol", [mask_invalid, apply_qa, scale_data]],
                    "_B02_60m.jp2": [(), "blue", [mask_invalid, apply_qa, scale_data]],
                    "_B03_60m.jp2": [(), "green", [mask_invalid, apply_qa, scale_data]],
                    "_B04_60m.jp2": [(), "red", [mask_invalid, apply_qa, scale_data]],
                    "_B05_60m.jp2": [(), "red_edge_703", [mask_invalid, apply_qa, scale_data]],
                    "_B06_60m.jp2": [(), "red_edge_740", [mask_invalid, apply_qa, scale_data]],
                    "_B07_60m.jp2": [(), "red_edge_782", [mask_invalid, apply_qa, scale_data]],
                    "_B8A_60m.jp2": [(), "nir", [mask_invalid, apply_qa, scale_data]],
                    "_B09_60m.jp2": [(), "narrow_nir", [mask_invalid, apply_qa, scale_data]],
                    "_B11_60m.jp2": [(), "swir1", [mask_invalid, apply_qa, scale_data]],
                    "_B12_60m.jp2": [(), "swir2", [mask_invalid, apply_qa, scale_data]],
                    "_SCL_60m.jp2": [(), "qa", []],
                },
    }

    req_dl_vars = {

        "S2MSI2A_R20m": {
            "coastal_aerosol":  ["_B01_20m.jp2", "_SCL_20m.jp2"],
            "blue":             ["_B02_20m.jp2", "_SCL_20m.jp2"],
            "green":            ["_B03_20m.jp2", "_SCL_20m.jp2"],
            "red":              ["_B04_20m.jp2", "_SCL_20m.jp2"],
            "red_edge_703":     ["_B05_20m.jp2", "_SCL_20m.jp2"],
            "red_edge_740":     ["_B06_20m.jp2", "_SCL_20m.jp2"],
            "red_edge_782":     ["_B07_20m.jp2", "_SCL_20m.jp2"],
            "nir":              ["_B8A_20m.jp2", "_SCL_20m.jp2"],
            "swir1":            ["_B11_20m.jp2", "_SCL_20m.jp2"],
            "swir2":            ["_B12_20m.jp2", "_SCL_20m.jp2"],
            "qa":               ["_SCL_20m.jp2"],
            "ndvi":             ["_B04_20m.jp2", "_B8A_20m.jp2", "_SCL_20m.jp2"],
            "mndwi":            ["_B03_20m.jp2", "_B11_20m.jp2", "_SCL_20m.jp2"],
            "vari_red_edge":    ["_B06_20m.jp2", "_B02_20m.jp2", "_B04_20m.jp2", "_SCL_20m.jp2"],
            "nmdi":             ["_B11_20m.jp2", "_B12_20m.jp2", "_B8A_20m.jp2", "_SCL_20m.jp2"],
            "psri":             ["_B02_20m.jp2", "_B04_20m.jp2", "_B06_20m.jp2", "_SCL_20m.jp2"],
            "bsi":              ["_B8A_20m.jp2", "_B11_20m.jp2", "_B04_20m.jp2", "_B02_20m.jp2", "_SCL_20m.jp2"],
            "r0":               ["_B02_20m.jp2", "_B03_20m.jp2", "_B04_20m.jp2", "_B8A_20m.jp2", "_SCL_20m.jp2"],
        },

        "S2MSI2A_R60m": {
            "coastal_aerosol":  ["_B01_60m.jp2", "_SCL_60m.jp2"],
            "blue":             ["_B02_60m.jp2", "_SCL_60m.jp2"],
            "green":            ["_B03_60m.jp2", "_SCL_60m.jp2"],
            "red":              ["_B04_60m.jp2", "_SCL_60m.jp2"],
            "red_edge_703":     ["_B05_60m.jp2", "_SCL_60m.jp2"],
            "red_edge_740":     ["_B06_60m.jp2", "_SCL_60m.jp2"],
            "red_edge_782":     ["_B07_60m.jp2", "_SCL_60m.jp2"],
            "nir":              ["_B8A_60m.jp2", "_SCL_60m.jp2"],
            "swir1":            ["_B11_60m.jp2", "_SCL_60m.jp2"],
            "swir2":            ["_B12_60m.jp2", "_SCL_60m.jp2"],
            "qa":               ["_SCL_60m.jp2"],
            "ndvi":             ["_B04_60m.jp2", "_B8A_60m.jp2", "_SCL_60m.jp2"],
            "mndwi":            ["_B03_60m.jp2", "_B11_60m.jp2", "_SCL_60m.jp2"],
            "vari_red_edge":    ["_B06_60m.jp2", "_B02_60m.jp2", "_B04_60m.jp2", "_SCL_60m.jp2"],
            "nmdi":             ["_B11_60m.jp2", "_B12_60m.jp2", "_B8A_60m.jp2", "_SCL_60m.jp2"],
            "psri":             ["_B02_60m.jp2", "_B04_60m.jp2", "_B06_60m.jp2", "_SCL_60m.jp2"],
            "bsi":              ["_B8A_60m.jp2", "_B11_60m.jp2", "_B04_60m.jp2", "_B02_60m.jp2", "_SCL_60m.jp2"],
            "r0":               ["_B02_60m.jp2", "_B03_60m.jp2", "_B04_60m.jp2", "_B8A_60m.jp2", "_SCL_60m.jp2"],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}

    return out

def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """
    
    post_processors = {
        "S2MSI2A_R20m": {
            "coastal_aerosol":  [],
            "blue":             [],
            "green":            [],
            "red":              [],
            "red_edge_703":     [],
            "red_edge_740":     [],
            "red_edge_782":     [],
            "nir":              [],
            "qa":               [],
            "swir1":            [],
            "swir2":            [],
            "psri":             [calc_psri],
            "ndvi":             [calc_normalized_difference],
            "nmdi":             [calc_nmdi],
            "vari_red_edge":    [calc_vari_red_egde],
            "bsi":              [calc_bsi],
            "mndwi":            [partial(calc_normalized_difference, bands = ["swir1", "green"])],
            "r0":               [calc_r0],
            },

        "S2MSI2A_R60m": {
            "coastal_aerosol":  [],
            "blue":             [],
            "green":            [],
            "red":              [],
            "red_edge_703":     [],
            "red_edge_740":     [],
            "red_edge_782":     [],
            "nir":              [],
            "qa":               [],
            "swir1":            [],
            "swir2":            [],
            "psri":             [calc_psri],
            "ndvi":             [calc_normalized_difference],
            "nmdi":             [calc_nmdi],
            "vari_red_edge":    [calc_vari_red_egde],
            "bsi":              [calc_bsi],
            "mndwi":            [partial(calc_normalized_difference, bands = ["swir1", "green"])],
            "r0":               [calc_r0],
            },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def time_func(fn):
    """Return a np.datetime64 given a filename.

    Parameters
    ----------
    fn : str
        Filename.

    Returns
    -------
    np.datetime64
        Date as described in the filename.
    """
    dtime = np.datetime64(dt.strptime(fn.split("_")[2], "%Y%m%dT%H%M%S"))
    return dtime

def s2_processor(scene_folder, variables):
    dss = [open_ds(glob.glob(os.path.join(scene_folder, "**", "*" + k), recursive = True)[0], decode_coords=None).isel(band=0).rename({"band_data": v[1]}) for k, v in variables.items()]
    ds = xr.merge(dss).drop_vars("band")
    return ds

def download(folder, latlim, lonlim, timelim, product_name, 
                req_vars, variables = None, post_processors = None, 
                extra_search_kwargs = {"cloudcoverpercentage": (0, 30)}):
    """Download SENTINEL2 data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.
    extra_search_kwargs : dict
        Extra search kwargs passed to SentinelAPI, by default {"cloudcoverpercentage": (0, 30)}.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """

    product_folder = os.path.join(folder, "SENTINEL2")

    appending = False
    fn = os.path.join(product_folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(product_folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]
            
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    bb = [lonlim[0], latlim[0], lonlim[1], latlim[1]]

    search_kwargs = {
                        "platformname": "Sentinel-2",
                        "producttype": "S2MSI2A",
                        # "limit": 10,
    }

    search_kwargs = {**search_kwargs, **extra_search_kwargs}

    def node_filter(node_info):
        fn = os.path.split(node_info["node_path"])[-1]
        to_dl = list(variables.keys())
        return np.any([x in fn for x in to_dl])
    # node_filter = None

    scenes = sentinelapi.download(product_folder, latlim, lonlim, timelim, search_kwargs, node_filter = node_filter)

    ds_new = sentinelapi.process_sentinel(scenes, variables, "SENTINEL2", time_func, os.path.split(fn)[-1], post_processors, bb = bb)

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(product_folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds

# if __name__ == "__main__":

#     folder = r"/Users/hmcoerver/Local/s2_test"
#     adjust_logger(True, folder, "INFO")
#     timelim = ["2022-03-29", "2022-04-25"]
#     latlim = [28.9, 29.7]
#     lonlim = [30.2, 31.2]

#     product_name = 'S2MSI2A_R60m'
#     req_vars = ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]
#     post_processors = None
#     variables = None
#     extra_search_kwargs = {"cloudcoverpercentage": (0, 30)}

#     ds = download(folder, latlim, lonlim, timelim, product_name, req_vars, 
#                 variables = None,  post_processors = None,
#                 extra_search_kwargs = extra_search_kwargs
#                  )



def pa_to_kpa(ds, var):
    """Convert Pa to kPa.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to adjust.
    var : str
        Variable in `ds` to adjust.

    Returns
    -------
    xr.Dataset
        Dataset in which `var` has been divided by 1000.
    """
    ds[var] = ds[var] / 1000
    ds[var].attrs = {}
    return ds


# -*- coding: utf-8 -*-
"""
Generates input data for `pywapor.se_root`.
"""
from pywapor.collect import downloader
from pywapor.general import compositer
from pywapor.enhancers.temperature import bt_to_lst
from pywapor.enhancers.other import drop_empty_times
import pywapor.general.levels as levels
from pywapor.general import aligner
import datetime
import os
import numpy as np
import pywapor.general.pre_defaults as defaults
from pywapor.general.logger import log

def rename_meteo(ds, *args):
    """Rename some variables in a dataset.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset whose variables will be renamed.

    Returns
    -------
    xr.Dataset
        Dataset with renamed variables.
    """
    renames = {
                "t_air": "t_air_i",
                "u2m": "u2m_i",
                "v2m": "v2m_i",
                "u": "u_i",
                "qv": "qv_i",
                "p_air": "p_air_i",
                "wv": "wv_i",
                "p_air_0": "p_air_0_i",
                "t_dew": "t_dew_i",
            }
    renames_filtered = {k: v for k,v in renames.items() if k in ds.data_vars}
    ds = ds.rename_vars(renames_filtered)
    return ds

def add_constants(ds, *args):
    """Adds default dimensionless constants to a dataset.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to which the constants will be added.

    Returns
    -------
    xr.Dataset
        Dataset with extra variables.
    """
    ds = ds.assign(defaults.constants_defaults())
    return ds

def main(folder, latlim, lonlim, timelim, sources = "level_1", bin_length = "DEKAD", enhancers = [], **kwargs):
    """Prepare input data for `se_root`.

    Parameters
    ----------
    folder : str
        Path to folder in which to store (intermediate) data.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    sources : str | dict
        Configuration for each variable and source.
    bin_length : int | "DEKAD"
        Composite length.
    enhancers : list | "default", optional
        Functions to apply to the xr.Dataset before creating the final
        output, by default "default".
    example_source : tuple, optional
        Which source to use for spatial alignment, overrides product selected
        through `sources`, by default None.

    Returns
    -------
    xr.Dataset
        Dataset with input data for `pywapor.se_root`.
    """
    t1 = datetime.datetime.now()
    log.info("> PRE_SE_ROOT").add()

    if not os.path.exists(folder):
        os.makedirs(folder)

    if isinstance(sources, str):
        sources = levels.pre_se_root_levels(sources)

    example_t_vars = [x for x in ["lst", "bt"] if x in sources.keys()]
    example_sources = {k:v for k,v in sources.items() if k in example_t_vars}
    other_sources = {k:v for k,v in sources.items() if k not in example_t_vars}

    general_enhancers = enhancers + [rename_meteo, add_constants, bt_to_lst, drop_empty_times]

    bins = compositer.time_bins(timelim, bin_length)
    adjusted_timelim = [bins[0], bins[-1]]
    buffered_timelim = [adjusted_timelim[0] - np.timedelta64(3, "D"), 
                        adjusted_timelim[1] + np.timedelta64(3, "D")]

    example_dss, example_sources = downloader.collect_sources(folder, example_sources, latlim, lonlim, adjusted_timelim, return_fps = False)

    if len(example_dss) == 0:
        lbl = f"Unable to collect the essential variable(s) (`{'and'.join(example_t_vars)}`) to which the other variables should be aligned."
        log.error("--> " + lbl)
        raise ValueError(lbl)
    
    other_dss, other_sources = downloader.collect_sources(folder, other_sources, latlim, lonlim, buffered_timelim, return_fps = False)
    dss= {**example_dss, **other_dss}

    ds = aligner.main(dss, sources, folder, general_enhancers, example_t_vars = example_t_vars)

    t2 = datetime.datetime.now()
    log.sub().info(f"< PRE_SE_ROOT ({str(t2 - t1)})")

    return ds

if __name__ == "__main__":

    # from pywapor.se_root import main as se_root

    # sources = "level_1"

    enhancers = []

    # folder = r"/Users/hmcoerver/pywapor_notebooks_2"
    # latlim = [28.9, 29.7]
    # lonlim = [30.2, 31.2]
    # timelim = [datetime.date(2021, 7, 1), datetime.date(2021, 7, 11)]
    # bin_length = "DEKAD"
    # example_source = None

    # # _ = adjust_logger(True, folder, "INFO")

    # sources = levels.pre_se_root_levels(sources)

    # sources["ndvi"]["products"] = [
    #     {'source': 'MODIS',
    #         'product_name': 'MOD13Q1.061',
    #         'enhancers': 'default'},
    #     {'source': 'MODIS', 
    #         'product_name': 'MYD13Q1.061', 
    #         'enhancers': 'default'},
    #     {'source': 'PROBAV',
    #         'product_name': 'S5_TOC_100_m_C1',
    #         'enhancers': 'default',
    #         'is_example': True}
    # ]

    # # example = levels.find_example(sources)

    # # print(example)
    # ds = main(folder, latlim, lonlim, timelim, sources, bin_length)
    
    # chunk_size = "20MiB"
    # ds = open_ds(r"/Users/hmcoerver/pywapor_notebooks_2/se_root_in.nc", chunk_size = chunk_size)
    # ds_out = se_root(ds)
    # ds_out = ds_out[["se_root"]]
    # save_ds(ds_out, r"/Users/hmcoerver/pywapor_notebooks_2/test_1.nc", chunk_size = chunk_size)

    # import xarray as xr
    # from dask.diagnostics import ProgressBar

    # ds = xr.open_dataset(r"/Users/hmcoerver/pywapor_notebooks_2/se_root_in.nc", 
    #                         chunks= {"time": -1, "x": 10,"y": 10})
    # ds_out = se_root(ds)
    # ds_out = ds_out[["se_root"]]

    # with ProgressBar():
    #     ds_out.to_netcdf(r"/Users/hmcoerver/pywapor_notebooks_2/test_3.nc")


import os
import pywapor.collect.protocol.sentinelapi as sentinelapi
from pywapor.general.curvilinear import regrid, create_grid
from pywapor.general.logger import log, adjust_logger
import glob
import xarray as xr
import numpy as np
from datetime import datetime as dt
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    variables = {
        "SL_2_LST___": {
                    "LST_in.nc": [(), "lst", []],
                    "geodetic_in.nc": [(), "coords", []],
        }
    }

    req_dl_vars = {
        "SL_2_LST___": {
            "lst": ["LST_in.nc", "geodetic_in.nc"]
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}

    return out

def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """
    post_processors = {
        "SL_2_LST___": {
            "lst": []
            },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def time_func(fn):
    """Return a np.datetime64 given a filename.

    Parameters
    ----------
    fn : str
        Filename.

    Returns
    -------
    np.datetime64
        Date as described in the filename.
    """
    start_dtime = np.datetime64(dt.strptime(fn.split("_")[7], "%Y%m%dT%H%M%S"))
    end_dtime = np.datetime64(dt.strptime(fn.split("_")[8], "%Y%m%dT%H%M%S"))
    dtime = start_dtime + (end_dtime - start_dtime)/2
    return dtime

def s3_processor(scene_folder, variables, bb = None, **kwargs):

    ncs = [glob.glob(os.path.join(scene_folder, "**", "*" + k), recursive = True)[0] for k in variables.keys()]

    ds = xr.open_mfdataset(ncs)

    ds = ds.set_coords(("longitude_in", "latitude_in"))
    ds = ds.rename_vars({"longitude_in": "x", "latitude_in": "y"})
    ds = ds.rename_dims({"rows": "ny", "columns": "nx"})
    ds = ds[["LST", "LST_uncertainty"]]

    ds = ds.where(ds.LST_uncertainty < 2.5)
    ds = ds.drop_vars("LST_uncertainty")

    grid_ds = create_grid(ds, 0.01, 0.01, bb = bb)
    ds = regrid(grid_ds, ds)
    ds = ds.rio.write_crs(4326)

    ds = ds.rename_vars({"LST": "lst"})

    return ds

def download(folder, latlim, lonlim, timelim, product_name, 
                req_vars, variables = None,  post_processors = None,
                extra_search_kwargs = {}):
    """Download SENTINEL3 data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.
    extra_search_kwargs : dict
        Extra search kwargs passed to SentinelAPI, by default {}.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    product_folder = os.path.join(folder, "SENTINEL3")

    appending = False
    fn = os.path.join(product_folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(product_folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    bb = [lonlim[0], latlim[0], lonlim[1], latlim[1]]

    search_kwargs = {
                        "platformname": "Sentinel-3",
                        "producttype": product_name,
                        # "limit": 10,
                        }

    search_kwargs = {**search_kwargs, **extra_search_kwargs}

    # NOTE node_filter not supported yet for Sentinel-3, so downloading all files for now.
    # https://github.com/sentinelsat/sentinelsat/issues/566
    scenes = sentinelapi.download(product_folder, latlim, lonlim, timelim, 
                                    search_kwargs, node_filter = None)

    ds_new = sentinelapi.process_sentinel(scenes, variables, "SENTINEL3", time_func, os.path.split(fn)[-1], post_processors, bb = bb)

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(product_folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds

if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Local/s3_test"
    adjust_logger(True, folder, "INFO")
    timelim = ["2022-03-25", "2022-04-15"]
    latlim = [29.4, 29.7]
    lonlim = [30.7, 31.0]

    product_name = 'SL_2_LST___'

    req_vars = ["lst"]
    post_processors = None
    variables = None
    extra_search_kwargs = {}

    ds = download(folder, latlim, lonlim, timelim, product_name, 
                req_vars, variables = variables,  post_processors = post_processors)





from pywapor.collect.protocol import opendap
from pywapor.collect.protocol.projections import get_crss
import os
import xarray as xr
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds
from pywapor.enhancers.temperature import kelvin_to_celsius
from functools import partial
import numpy as np
from pywapor.enhancers.pressure import pa_to_kpa

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {
        "inst3_2d_asm_Nx": {
                    "t2m": [("time", "lat", "lon"), "t_air"],
                    "u2m": [("time", "lat", "lon"), "u2m"],
                    "v2m": [("time", "lat", "lon"), "v2m"],
                    "qv2m": [("time", "lat", "lon"), "qv"],
                    "tqv": [("time", "lat", "lon"), "wv"],
                    "ps": [("time", "lat", "lon"), "p_air"],
                    "slp": [("time", "lat", "lon"), "p_air_0"],
                        },
    }

    req_dl_vars = {
        "inst3_2d_asm_Nx": {
            "t_air": ["t2m"],
            "t_air_max": ["t2m"],
            "t_air_min": ["t2m"],
            "u2m": ["u2m"],
            "v2m": ["v2m"],
            "qv": ["qv2m"],
            "wv": ["tqv"],
            "p_air": ["ps"],
            "p_air_0": ["slp"],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}

    return out

def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "inst3_2d_asm_Nx": {
            "t_air": [kelvin_to_celsius], 
            "t_air_max": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_max")],
            "t_air_min": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_min")],
            "u2m": [],
            "v2m": [],
            "qv": [],
            "wv": [],
            "p_air": [pa_to_kpa],
            "p_air_0": [pa_to_kpa],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def download(folder, latlim, lonlim, timelim, product_name, req_vars,
                 variables = None, post_processors = None):
    """Download GEOS5 data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "GEOS5")

    appending = False
    fn = os.path.join(folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]


    spatial_buffer = True
    if spatial_buffer:
        latlim = [latlim[0] - 0.25, latlim[1] + 0.25]
        lonlim = [lonlim[0] - 0.3125, lonlim[1] + 0.3125]

    coords = {"x": ["lon", lonlim], "y": ["lat", latlim], "t": ["time", timelim]}

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    data_source_crs = get_crss("WGS84")

    url = f"https://opendap.nccs.nasa.gov/dods/GEOS-5/fp/0.25_deg/assim/{product_name}"

    timedelta = np.timedelta64(90, "m")

    ds_new = opendap.download_xarray(url, fn, coords, variables, post_processors, 
                                    data_source_crs = data_source_crs,
                                    timedelta = timedelta)

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds

if __name__ == "__main__":

    import datetime

    folder = r"/Users/hmcoerver/Downloads/pywapor_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 11)]

    # GEOS5.
    product_name = "inst3_2d_asm_Nx"

    variables = None
    post_processors = None

    req_vars = [
                "t_air", "u2m", "v2m", "qv", 
                "wv", 
                "p_air", "p_air_0"
                ]

    ds = download(folder, latlim, lonlim, timelim, product_name, req_vars = req_vars)
    print(ds.rio.crs, ds.rio.grid_mapping)


import datetime
import pandas as pd
import pywapor.collect.accounts as accounts
from pywapor.general.logger import log
from pywapor.collect.protocol.projections import get_crss
import pywapor.collect.protocol.opendap as opendap
import fnmatch
import os
import numpy as np
from functools import partial
import xarray as xr
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds
from pywapor.collect.protocol.crawler import find_paths
from pywapor.enhancers.temperature import kelvin_to_celsius
from pywapor.enhancers.pressure import pa_to_kpa

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    variables = {
        "M2I1NXASM.5.12.4": {
                    "T2M": [("time", "lat", "lon"), "t_air"],
                    "U2M": [("time", "lat", "lon"), "u2m"],
                    "V2M": [("time", "lat", "lon"), "v2m"],
                    "QV2M": [("time", "lat", "lon"), "qv"],
                    "TQV": [("time", "lat", "lon"), "wv"],
                    "PS": [("time", "lat", "lon"), "p_air"],
                    "SLP": [("time", "lat", "lon"), "p_air_0"],
                        },
        "M2T1NXRAD.5.12.4": {
                    "SWGNT": [("time", "lat", "lon"), "ra"],
                        }
    }

    req_dl_vars = {
        "M2I1NXASM.5.12.4": {
            "t_air": ["T2M"],
            "t_air_max" :["T2M"],
            "t_air_min" :["T2M"],
            "u2m": ["U2M"],
            "v2m": ["V2M"],
            "qv": ["QV2M"],
            "wv": ["TQV"],
            "p_air": ["PS"],
            "p_air_0": ["SLP"],
        },
        "M2T1NXRAD.5.12.4": {
            "ra": ["SWGNT"],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "M2I1NXASM.5.12.4": {
            "t_air": [kelvin_to_celsius], 
            "t_air_max": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_max")],
            "t_air_min": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_min")],
            "u2m": [],
            "v2m": [],
            "qv": [],
            "wv": [],
            "p_air": [pa_to_kpa],
            "p_air_0": [pa_to_kpa],
        },
        "M2T1NXRAD.5.12.4": {
            "ra": [],
        },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def fn_func(product_name, tile):
    """Returns a client-side filename at which to store data.

    Parameters
    ----------
    product_name : str
        Name of the product to download.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        Filename.
    """
    fn = f"{product_name}_{tile.strftime('%Y%m%d')}.nc"
    return fn

def url_func(product_name, tile):
    """Returns a url at which to collect MERRA2 data.

    Parameters
    ----------
    product_name : str
        Name of the product to download.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        The url.
    """
    def _filter(tag):
        tag_value = tag["href"]
        if tag_value[-5:] == ".html":
            tag_value = tag_value[:-5] 
        return tag_value
    # Find the existing tiles for the given year and month, this is necessary
    # because the version number (`\d{3}`) in the filenames is irregular.
    regex = r"MERRA2_\d{3}\..*\.\d{8}.nc4.html"
    url = f"https://goldsmr4.gesdisc.eosdis.nasa.gov/opendap/MERRA2/{product_name}/{tile.year}/{tile.month:02d}/contents.html"
    tile_names = find_paths(url, regex, filter = _filter)

    # Find which of the existing tiles matches with the date.
    fn_pattern = f"MERRA2_*.*.{tile.strftime('%Y%m%d')}.nc4"
    fn = fnmatch.filter(tile_names, fn_pattern)[0]

    # Create the final url.
    url = f"https://goldsmr4.gesdisc.eosdis.nasa.gov/opendap/MERRA2/{product_name}/{tile.year}/{tile.month:02d}/{fn}.nc4?"  
    return url

def download(folder, latlim, lonlim, timelim, product_name, req_vars,
                variables = None, post_processors = None):
    """Download MERRA2 data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    
    folder = os.path.join(folder, "MERRA2")
    appending = False
    fn = os.path.join(folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]

    spatial_buffer = True
    if spatial_buffer:
        latlim = [latlim[0] - 0.5, latlim[1] + 0.5]
        lonlim = [lonlim[0] - 0.625, lonlim[1] + 0.625]

    tiles = pd.date_range(timelim[0], timelim[1], freq="D")
    coords = {"x": ["lon", lonlim], "y":["lat", latlim], "t": ["time", timelim]}
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)
    
    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    timedelta = np.timedelta64(30, "m")
    data_source_crs = get_crss("WGS84")
    parallel = False
    spatial_tiles = False
    un_pw = accounts.get("NASA")
    request_dims = True

    ds_new = opendap.download(fn, product_name, coords, 
                variables, post_processors, fn_func, url_func, un_pw = un_pw, 
                tiles = tiles, data_source_crs = data_source_crs, parallel = parallel, 
                spatial_tiles = spatial_tiles, request_dims = request_dims,
                timedelta = timedelta)

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds


if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Downloads/pywapor_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 11)]

    variables = None
    post_processors = None
    
    # # MERRA2.
    # wanted = [
    #             ("M2I1NXASM.5.12.4", ["t_air", "u2m", "v2m", "qv", "wv", "p_air", "p_air_0"]),
    #             # ("M2T1NXRAD.5.12.4", ["ra"]),
    #         ]

    product_name = "M2I1NXASM.5.12.4"
    req_vars = ["t_air", "u2m", "v2m", "qv", "wv", "p_air", "p_air_0"]

    # for product_name, req_vars in wanted:
    # ds = download(folder, latlim, lonlim, timelim, product_name, req_vars)
    # print(ds.rio.crs, ds.rio.grid_mapping)



"""This module defines the default products to use for each variable per level
and defines the variables and their units.
"""
import xarray as xr
import numpy as np
import copy

def get_source_level_selections():
    """Gives a definition of which products are used for each variable per level.

    Returns
    -------
    dict
        Dictionary with the products to use per variable per level.
    """

    source_selection_level2 = {
        "ndvi": ["PROBAV"],
        "r0": ["PROBAV"],
        "lst": ["MOD11", "MYD11"],
        "lulc": ["WAPOR"],
        "z": ["SRTM"],
        "p_24": ["CHIRPS"],
        "ra_24": ["MERRA2"],
        't_air_24': ["GEOS5"],
        't_air_min_24': ["GEOS5"], 
        't_air_max_24': ["GEOS5"],
        'u2m_24': ["GEOS5"],
        'v2m_24': ["GEOS5"],
        'p_air_0_24': ["GEOS5"],
        'qv_24': ["GEOS5"],
        "t_air_i": ["GEOS5"],
        "u2m_i": ["GEOS5"],
        "v2m_i": ["GEOS5"],
        "qv_i": ["GEOS5"],
        "wv_i": ["GEOS5"],
        "p_air_i": ["GEOS5"],
        "p_air_0_i": ["GEOS5"],
        "lw_offset": ["STATICS"],
        "lw_slope": ["STATICS"],
        "r0_bare": ["STATICS"],
        "r0_full": ["STATICS"],
        "rn_offset": ["STATICS"],
        "rn_slope": ["STATICS"],
        "t_amp_year": ["STATICS"],
        "t_opt": ["STATICS"],
        "vpd_slope": ["STATICS"],
        "z_oro": ["STATICS"],
        # "z_obst_max": ["STATICS"],
        # "rs_min": ["STATICS"],
        # "land_mask": ["STATICS"],
    }

    source_selection_level1 = {
        "ndvi": ["MOD13", "MYD13"],
        "r0": ["MCD43"],
        "lst": ["MOD11", "MYD11"],
        "lulc": ["WAPOR"],
        "z": ["SRTM"],
        "p_24": ["CHIRPS"],
        "ra_24": ["MERRA2"],
        't_air_24': ["GEOS5"],
        't_air_min_24': ["GEOS5"], 
        't_air_max_24': ["GEOS5"],
        'u2m_24': ["GEOS5"],
        'v2m_24': ["GEOS5"],
        'p_air_0_24': ["GEOS5"],
        'qv_24': ["GEOS5"],
        "t_air_i": ["GEOS5"],
        "u2m_i": ["GEOS5"],
        "v2m_i": ["GEOS5"],
        "qv_i": ["GEOS5"],
        "wv_i": ["GEOS5"],
        "p_air_i": ["GEOS5"],
        "p_air_0_i": ["GEOS5"],
        "lw_offset": ["STATICS"],
        "lw_slope": ["STATICS"],
        "r0_bare": ["STATICS"],
        "r0_full": ["STATICS"],
        "rn_offset": ["STATICS"],
        "rn_slope": ["STATICS"],
        "t_amp_year": ["STATICS"],
        "t_opt": ["STATICS"],
        "vpd_slope": ["STATICS"],
        "z_oro": ["STATICS"],
        # "z_obst_max": ["STATICS"],
        # "rs_min": ["STATICS"],
        # "land_mask": ["STATICS"],
    }

    levels = {"level_1": source_selection_level1, 
              "level_2": source_selection_level2}

    return levels

def fill_attrs(ds, *args):
    """Fills in the attributes of a xr.Dataset.

    Parameters
    ----------
    ds : xr.Dataset
        The dataset in which to add metadata attributes.

    Returns
    -------
    xr.Dataset
        The dataset with extra metadata.
    """

    defs = get_var_definitions()

    for var in list(ds.variables):

        if var in defs.keys():

            attributes = defs[var]

            _ = attributes.pop("definition")

            for k, v in attributes.items():
                ds[var].attrs[k] = v

    ds = fill_source_data_attr(ds)

    return ds

def initiate_ds(ds = None, dummy_data = False):
    """Create a template input xr.Dataset for et_look.

    Parameters
    ----------
    ds : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """

    if isinstance(ds, type(None)):
        ds = xr.Dataset()

    defs = get_var_definitions()

    for var, attributes in defs.items():

        if not var in list(ds.variables):
            if dummy_data:
                ds[var] = 0.5
            else:
                ds[var] = None

        _ = attributes.pop("definition")

        ds[var].attrs = attributes

    return ds

def flatten(list_of_lists):
    flat_list = list()
    def _flatten(list_of_lists):
        for item in list_of_lists:
            if type(item) == list:
                _flatten(item)
            else:
                flat_list.append(item)
        return flat_list
    _flatten(list_of_lists)
    return flat_list

def iterate(depends_on):
    j = 0
    depends_on_n = copy.deepcopy(depends_on)
    depends_on_n1 = dict()
    while depends_on_n != depends_on_n1:
        j += 1
        # print(f"{j}")
        depends_on_n1 = copy.deepcopy(depends_on_n)
        for var in depends_on_n.keys():
            for i, dependant in enumerate(depends_on_n[var]):
                if dependant in depends_on_n.keys():
                    depends_on_n[var][i] = depends_on[dependant]
        for var in depends_on_n.keys():
            depends_on_n[var] = np.unique(flatten(depends_on_n[var])).tolist()
        if j > 100:
            break
    return depends_on_n

def fill_source_data_attr(ds):
    """Determines the input data sources each variable depends on, based
    on the "calculated_with" attribute in the `ds`.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset for which to determine the source data.

    Returns
    -------
    xr.Dataset
        Dataset with new attribute information.
    """
    depends_on = {y: flatten([ds[y].calculated_with]) for y in ds.variables if "calculated_with" in ds[y].attrs.keys()}
    depends_on_n = iterate(depends_on)
    for var, dependants in depends_on_n.items():
        if var in list(ds.variables):
            ds[var].attrs["source_data"] = dependants
    return ds

def get_var_definitions():
    """Gives he defintion and unit of all the variables used in `et_look` and
    `se_root`.

    Returns
    -------
    dict
        Keys are variable names as used in `et_look` and `se_root`, values are
        dictionaries with keys `long_name`, `units` and `definition`.
    """

    defs = {
        "IO": {
            "long_name": "Solar constant",
            "units": "W m-2",
            "definition": ""
        },
        "B0c": {
            "long_name": "Beam irradiance normal to the solar beam",
            "units": "W m-2",
            "definition": ""
        },
        "Bhc": {
            "long_name": "Beam irradiance at a horizontal surface",
            "units": "W m-2",
            "definition": ""
        },
        "Dhc": {
            "long_name": "Diffuse irradiance at a horizontal surface",
            "units": "W m-2",
            "definition": ""
        },
        "G0": {
            "long_name": "Ext rad normal to solar beam",
            "units": "W m-2",
            "definition": ""
        },
        "L_bare": {
            "long_name": "Monin obukhov length dry vegetation",
            "units": "m",
            "definition": ""
        },
        "L_full": {
            "long_name": "Monin obukhov length wet vegetation",
            "units": "m",
            "definition": ""
        },
        "Tl2": {
            "long_name": "Airmass 2 linke atmospheric turbidity factor",
            "units": "-",
            "definition": ""
        },
        "ad_24": {
            "long_name": "Daily air density",
            "units": "kg m-3",
            "definition": ""
        },
        "ad_dry_24": {
            "long_name": "Daily dry air density",
            "units": "kg m-3",
            "definition": ""
        },
        "ad_dry_i": {
            "long_name": "Instantaneous dry air density",
            "units": "kg m-3",
            "definition": ""
        },
        "ad_i": {
            "long_name": "Instantaneous air density",
            "units": "kg m-3",
            "definition": ""
        },
        "ad_moist_24": {
            "long_name": "Daily moist air density",
            "units": "kg m-3",
            "definition": ""
        },
        "ad_moist_i": {
            "long_name": "Instantaneous moist air density",
            "units": "kg m-3",
            "definition": ""
        },
        "angle": {
            "long_name": "",
            "units": "",
            "definition": ""
        },
        "aspect": {
            "long_name": "Aspect (0 is north; pi is south)",
            "units": "radians",
            "definition": ""
        },
        "band": {
            "long_name": "",
            "units": "",
            "definition": ""
        },
        "day_angle": {
            "long_name": "Day angle",
            "units": "radians",
            "definition": ""
        },
        "dd": {
            "long_name": "Damping depth",
            "units": "m",
            "definition": ""
        },
        "decl": {
            "long_name": "Solar declination",
            "units": "radians",
            "definition": ""
        },
        "disp": {
            "long_name": "Displacement height",
            "units": "m",
            "definition": ""
        },
        "disp_bare": {
            "long_name": "Displacement height (bare)",
            "units": "m",
            "definition": ""
        },
        "disp_full": {
            "long_name": "Displacement height (fully vegetated)",
            "units": "m",
            "definition": ""
        },
        "doy": {
            "long_name": "Day of year",
            "units": "-",
            "definition": ""
        },
        "dtime": {
            "long_name": "Decimal time",
            "units": "hours",
            "definition": ""
        },
        "e_24": {
            "long_name": "Daily evaporation energy equivalent",
            "units": "W m-2",
            "definition": ""
        },
        "e_24_init": {
            "long_name": "Initial estimate radiation equivalent daily evaporation",
            "units": "W m-2",
            "definition": ""
        },
        "e_24_mm": {
            "long_name": "Daily evaporation in mm",
            "units": "mm day-1",
            "definition": ""
        },
        "emiss_atm_i": {
            "long_name": "Instantaneous atmospheric emissivity",
            "units": "-",
            "definition": ""
        },
        # "epoch": {
        #     "long_name": "",
        #     "units": "",
        #     "definition": ""
        # },
        # "epoch_ends": {
        #     "long_name": "",
        #     "units": "",
        #     "definition": ""
        # },
        # "epoch_starts": {
        #     "long_name": "",
        #     "units": "",
        #     "definition": ""
        # },
        "et_24_mm": {
            "long_name": "Daily evapotranspiration in mm",
            "units": "mm day-1",
            "definition": ""
        },
        "et_ref_24": {
            "long_name": "Daily reference evapotranspiration (well watered grass) energy equivalent",
            "units": "W m-2",
            "definition": ""
        },
        "et_ref_24_mm": {
            "long_name": "Reference evapotranspiration (well watered grass)",
            "units": "mm day-1",
            "definition": ""
        },
        "g0_24": {
            "long_name": "Daily soil heat flux",
            "units": "W m-2",
            "definition": ""
        },
        "g0_bs": {
            "long_name": "Bare soil heat flux",
            "units": "W m-2",
            "definition": ""
        },
        "h0": {
            "long_name": "Solar elevation angle",
            "units": "degrees",
            "definition": ""
        },
        "h0ref": {
            "long_name": "Solar elevation angle corrected for refraction",
            "units": "degrees",
            "definition": ""
        },
        "h_bare": {
            "long_name": "Sensible heat flux for dry bare soil",
            "units": "W m-2",
            "definition": ""
        },
        "h_canopy_24_init": {
            "long_name": "Initial estimate of the sensible heat flux",
            "units": "W m-2",
            "definition": ""
        },
        "h_full": {
            "long_name": "Sensible heat flux full vegetation",
            "units": "W m-2",
            "definition": ""
        },
        "h_soil_24_init": {
            "long_name": "Initial estimate of the sensible heat flux for soil",
            "units": "W m-2",
            "definition": ""
        },
        "ha": {
            "long_name": "Solar hour angle",
            "units": "radians",
            "definition": ""
        },
        "ied": { # TODO check this with iesd
            "long_name": "Inverse earth sun distance",
            "units": "AU",
            "definition": ""
        },
        "iesd": {
            "long_name": "Inverse earth sun distance",
            "units": "AU",
            "definition": ""
        },
        "int_mm": {
            "long_name": "Interception",
            "units": "mm day-1",
            "definition": ""
        },
        "int_wm2": {
            "long_name": "Interception",
            "units": "W m-2",
            "definition": ""
        },
        "l_net": {
            "long_name": "Daily net longwave radiation",
            "units": "W m-2",
            "definition": ""
        },
        "lai": {
            "long_name": "Leaf area index",
            "units": "-",
            "definition": ""
        },
        "lai_eff": {
            "long_name": "Effective leaf area index",
            "units": "-",
            "definition": ""
        },
        "land_mask": {
            "long_name": "Land use classification",
            "units": "-",
            "definition": ""
        },
        "lat": {
            "long_name": "Latitude",
            "units": "degrees",
            "definition": ""
        },
        "lat_deg": {
            "long_name": "Latitude in degrees",
            "units": "degrees",
            "definition": ""
        },
        "lat_rad": {
            "long_name": "",
            "units": "",
            "definition": ""
        },
        "lh_24": {
            "long_name": "Daily latent heat of evaporation",
            "units": "J kg-1",
            "definition": ""
        },
        "lon": {
            "long_name": "Longitude",
            "units": "degrees",
            "definition": ""
        },
        "lon_deg": {
            "long_name": "Longitude in degrees",
            "units": "degrees",
            "definition": ""
        },
        "lon_rad": {
            "long_name": "Longitude in radians",
            "units": "radians",
            "definition": ""
        },
        "lst": {
            "long_name": "Land surface temperature",
            "units": "K",
            "definition": ""
        },
        "lst_max": {
            "long_name": "Maximum temperature at dry conditions",
            "units": "K",
            "definition": ""
        },
        "lst_min": {
            "long_name": "Minimum temperature at wet conditions",
            "units": "K",
            "definition": ""
        },
        "biomass_prod": {
            "long_name": "",
            "units": "kgDM ha-1 day-1",
            "definition": ""
        },
        "npp": {
            "long_name": "Net primary production",
            "units": "gC m-2 day-1",
            "definition": ""
        },
        "apar": {
            "long_name": "Absorbed photosynthetically active radiation",
            "units": "MJ m-2 day-1",
            "definition": ""
        },    
        "fpar": {
            "long_name": "Fraction of photosynthetically active radiation",
            "units": "-",
            "definition": ""
        },
        "lue": {
            "long_name": "Light use efficiency",
            "units": "gDM MJ-1",
            "definition": ""
        },
        "lue_max": {
            "long_name": "Maximum light use efficiency",
            "units": "gDM MJ-1",
            "definition": ""
        },
        "lw_offset": {
            "long_name": "Offset of the tau-term in the FAO-56 longwave radiation relationship",
            "units": "-",
            "definition": ""
        },
        "lw_slope": {
            "long_name": "Slope of the tau-term in the FAO-56 longwave radiation relationship",
            "units": "-",
            "definition": ""
        },
        "m": {
            "long_name": "Relative optical airmass",
            "units": "-",
            "definition": ""
        },
        "ndvi": {
            "long_name": "Normalized difference vegetation index",
            "units": "-",
            "definition": ""
        },
        "p_24": {
            "long_name": "Daily precipitation",
            "units": "mm day-1",
            "definition": ""
        },
        "p_air_0_24": {
            "long_name": "Daily air pressure at sea level",
            "units": "kpa",
            "definition": ""
        },
        "p_air_0_24_mbar": {
            "long_name": "Daily air pressure at sea level",
            "units": "mbar",
            "definition": ""
        },
        "p_air_0_i": {
            "long_name": "Instantaneous air pressure at sea level",
            "units": "-",
            "definition": ""
        },
        "p_air_24": {
            "long_name": "Daily air pressure",
            "units": "mbar",
            "definition": ""
        },
        "p_air_i": {
            "long_name": "Instantaneous air pressure",
            "units": "mbar",
            "definition": ""
        },
        "psy_24": {
            "long_name": "Daily psychrometric constant",
            "units": "mbar K-1",
            "definition": ""
        },
        "quantile": {
            "long_name": "",
            "units": "",
            "definition": ""
        },
        "qv_24": {
            "long_name": "Daily specific humidity",
            "units": "kg kg-1",
            "definition": ""
        },
        "qv_i": {
            "long_name": "Instantaneous specific humidity",
            "units": "kg kg-1",
            "definition": ""
        },
        "r0": {
            "long_name": "Albedo",
            "units": "-",
            "definition": ""
        },
        "r0_bare": {
            "long_name": "Dry bare soil surface albedo",
            "units": "-",
            "definition": ""
        },
        "r0_full": {
            "long_name": "Surface albedo under full vegetation cover",
            "units": "-",
            "definition": ""
        },
        "r_canopy": {
            "long_name": "Canopy resistance",
            "units": "s m-1",
            "definition": ""
        },
        "r_canopy_unstressed": {
            'long_name': 'Canopy resistance for zero moisture stress', 
            'units': 's m-1', 
            'definition': ''},
        "r_canopy_0": {
            "long_name": "Atmospheric canopy resistance",
            "units": "s m-1",
            "definition": ""
        },
        "r_soil": {
            "long_name": "Soil resistance",
            "units": "s m-1",
            "definition": ""
        },
        "ra_24": {
            "long_name": "Daily solar radiation",
            "units": "W m-2",
            "definition": ""
        },
        "ra_24_toa_flat": {
            "long_name": "Daily solar radiation at the top of atmosphere for a flat surface",
            "units": "W m-2",
            "definition": ""
        },
        "ra_canopy_init": {
            "long_name": "Initial canopy aerodynamic resistance",
            "units": "s m-1",
            "definition": ""
        },
        "ra_hor_clear_i": {
            "long_name": "Total clear-sky irradiance on a horizontal surface",
            "units": "W m-2",
            "definition": ""
        },
        "ra_soil_init": {
            "long_name": "Initial soil aerodynamic resistance",
            "units": "s m-1",
            "definition": ""
        },
        "raa": {
            "long_name": "Aerodynamical resistance dry surface",
            "units": "s m-1",
            "definition": ""
        },
        "rac": {
            "long_name": "Aerodynamical resistance canopy",
            "units": "s m-1",
            "definition": ""
        },
        "ras": {
            "long_name": "Aerodynamical resistance",
            "units": "s m-1",
            "definition": ""
        },
        "rn_24": {
            "long_name": "Daily net radiation",
            "units": "W m-2",
            "definition": ""
        },
        "rn_24_canopy": {
            "long_name": "Daily net radiation for the canopy",
            "units": "W m-2",
            "definition": ""
        },
        "rn_24_grass": {
            "long_name": "Daily net radiation for reference grass",
            "units": "W m-2",
            "definition": ""
        },
        "rn_24_soil": {
            "long_name": "Daily net radiation for soil",
            "units": "W m-2",
            "definition": ""
        },
        "rn_bare": {
            "long_name": "Net radiation bare soil",
            "units": "W m-2",
            "definition": ""
        },
        "rn_full": {
            "long_name": "Net radiation full vegetation",
            "units": "W m-2",
            "definition": ""
        },
        "rn_offset": {
            "long_name": "Offset rn/g0-relation water",
            "units": "-",
            "definition": ""
        },
        "rn_slope": {
            "long_name": "Slope rn/g0-relation water",
            "units": "-",
            "definition": ""
        },
        "rotm": {
            "long_name": "Rayleigh optical thickness at airmass m",
            "units": "-",
            "definition": ""
        },
        "rs_min": {
            "long_name": "Minimal stomatal resistance",
            "units": "s m-1",
            "definition": ""
        },
        "sc": {
            "long_name": "Seasonal correction",
            "units": "hours",
            "definition": ""
        },
        "se_root": {
            "long_name": "Relative root zone soil moisture",
            "units": "-",
            "definition": ""
        },
        "sf_soil": {
            "long_name": "Soil fraction",
            "units": "-",
            "definition": ""
        },
        "slope": {
            "long_name": "Slope",
            "units": "radians",
            "definition": ""
        },
        # "sources": {
        #     "long_name": "",
        #     "units": "",
        #     "definition": ""
        # },
        # "spatial_ref": {
        #     "long_name": "",
        #     "units": "",
        #     "definition": ""
        # },
        "ssvp_24": {
            "long_name": "Daily slope of saturated vapour pressure curve",
            "units": "mbar K-1",
            "definition": ""
        },
        "stc": {
            "long_name": "Soil thermal conductivity",
            "units": "W m-1 K-1",
            "definition": ""
        },
        "stress_moist": {
            "long_name": "Stress factor for root zone soil moisture",
            "units": "-",
            "definition": ""
        },
        "stress_rad": {
            "long_name": "Stress factor for radiation",
            "units": "-",
            "definition": ""
        },
        "stress_temp": {
            "long_name": "Stress factor for air temperature",
            "units": "-",
            "definition": ""
        },
        "stress_vpd": {
            "long_name": "Stress factor for vapour pressure deficit",
            "units": "-",
            "definition": ""
        },
        "svp_24": {
            "long_name": "Daily saturated vapour pressure",
            "units": "mbar",
            "definition": ""
        },
        "t_24": {
            "long_name": "Daily transpiration energy equivalent",
            "units": "W m-2",
            "definition": ""
        },
        "t_24_init": {
            "long_name": "Initial estimate radiation equivalent daily transpiration",
            "units": "W m-2",
            "definition": ""
        },
        "t_24_mm": {
            "long_name": "Daily transpiration in mm",
            "units": "mm day-1",
            "definition": ""
        },
        "t_air_24": {
            "long_name": "Daily air temperature",
            "units": "C",
            "definition": ""
        },
        "t_air_i": {
            "long_name": "Instantaneous air temperature",
            "units": "C",
            "definition": ""
        },
        "t_air_k_24": {
            "long_name": "Daily air temperature in kelvin",
            "units": "K",
            "definition": ""
        },
        "t_air_k_i": {
            "long_name": "Instantaneous air temperature in kelvin",
            "units": "K",
            "definition": ""
        },
        "t_air_max_24": {
            "long_name": "Daily maximum temperature",
            "units": "C",
            "definition": ""
        },
        "t_air_min_24": {
            "long_name": "Daily minimum temperature",
            "units": "C",
            "definition": ""
        },
        "t_amp_year": {
            "long_name": "Yearly air temperature amplitude",
            "units": "K",
            "definition": ""
        },
        "t_diff": {
            "long_name": "Lapse-rate temperature adjustment",
            "units": "C",
            "definition": ""
        },
        "t_max_bare": {
            "long_name": "Maximum temperature at bare soil",
            "units": "K",
            "definition": ""
        },
        "t_max_full": {
            "long_name": "Maximum temperature at full vegetation cover",
            "units": "K",
            "definition": ""
        },
        "t_opt": {
            "long_name": "Optimum air temperature for plant growth",
            "units": "C",
            "definition": ""
        },
        "t_wet_i": {
            "long_name": "Instantaneous wet bulb temperature",
            "units": "C",
            "definition": ""
        },
        "t_wet_k_i": {
            "long_name": "Instantaneous wet bulb temperature in kelvin",
            "units": "K",
            "definition": ""
        },
        # "time": {
        #     "long_name": "",
        #     "units": "",
        #     "definition": ""
        # },
        "trans_24": {
            "long_name": "Daily atmospheric transmissivity",
            "units": "-",
            "definition": ""
        },
        "u2m_24": {
            "long_name": "Daily average eastward wind speed at 2 meter",
            "units": "m s-1",
            "definition": ""
        },
        "u2m_i": {
            "long_name": "Instantaneous eastward wind speed at 2 meter",
            "units": "m s-1",
            "definition": ""
        },
        "u_24": {
            "long_name": "Daily wind speed at observation height",
            "units": "m s-1",
            "definition": ""
        },
        "u_b_24": {
            "long_name": "Daily wind speed at blending height",
            "units": "m s-1",
            "definition": ""
        },
        "u_b_i_bare": {
            "long_name": "Instantaneous wind speed at blending height for bare soil",
            "units": "m s-1",
            "definition": ""
        },
        "u_b_i_full": {
            "long_name": "Instantaneous wind speed at blending height for full vegetation",
            "units": "m s-1",
            "definition": ""
        },
        "u_i": {
            "long_name": "Instantaneous wind speed at observation height",
            "units": "m s-1",
            "definition": ""
        },
        "u_i_soil": {
            "long_name": "Instantaneous wind speed just above soil surface",
            "units": "m s-1",
            "definition": ""
        },
        "u_star_24_init": {
            "long_name": "Initial estimate of the daily friction velocity",
            "units": "m s-1",
            "definition": ""
        },
        "u_star_24_soil_init": {
            "long_name": "Initial estimate of the daily friction velocity for soil",
            "units": "m s-1",
            "definition": ""
        },
        "u_star_i_bare": {
            "long_name": "Instantaneous friction velocity bare soil",
            "units": "m s-1",
            "definition": ""
        },
        "u_star_i_full": {
            "long_name": "Instantaneous friction velocity vegetation",
            "units": "m s-1",
            "definition": ""
        },
        "v2m_24": {
            "long_name": "Daily average northward wind speed at 2 meter",
            "units": "m s-1",
            "definition": ""
        },
        "v2m_i": {
            "long_name": "Instantaneous northward wind speed at 2 meter",
            "units": "m s-1",
            "definition": ""
        },
        "vc": {
            "long_name": "Vegetation cover",
            "units": "-",
            "definition": ""
        },
        "vhc": {
            "long_name": "Volumetric heat capacity",
            "units": "J m-3 K-1",
            "definition": ""
        },
        "vp_24": {
            "long_name": "Daily vapour pressure",
            "units": "mbar",
            "definition": ""
        },
        "vp_i": {
            "long_name": "Instantaneous vapour pressure",
            "units": "mbar",
            "definition": ""
        },
        "vpd_24": {
            "long_name": "Daily vapour pressure deficit",
            "units": "mbar",
            "definition": ""
        },
        "vpd_slope": {
            "long_name": "Vapour pressure stress curve slope",
            "units": "mbar-1",
            "definition": ""
        },
        "ws": {
            "long_name": "Sunset hour angle",
            "units": "radians",
            "definition": ""
        },
        "wv_i": {
            "long_name": "Instantaneous total column atmospheric water vapor",
            "units": "kg m-2",
            "definition": ""
        },
        "z": {
            "long_name": "Elevation above sea level",
            "units": "m",
            "definition": ""
        },
        "z0m": {
            "long_name": "Surface roughness",
            "units": "m",
            "definition": ""
        },
        "z0m_bare": {
            "long_name": "Surface roughness (bare soil)",
            "units": "m",
            "definition": ""
        },
        "z0m_full": {
            "long_name": "Surface roughness (fully vegetated soil)",
            "units": "m",
            "definition": ""
        },
        "aod550_i": {
            "long_name": "",
            "units": "",
            "definition": ""
        },
        "fraction_h_bare": {
            "long_name": "",
            "units": "",
            "definition": ""
        },
        "fraction_h_full": {
            "long_name": "",
            "units": "",
            "definition": ""
        },
        "z_obst": {
            "long_name": "Obstacle height",
            "units": "m",
            "definition": ""
        },
        "z_obst_max": {
            "long_name": "Maximum obstacle height",
            "units": "m",
            "definition": ""
        },
        "z_oro": {
            "long_name": "Orographic roughness",
            "units": "m",
            "definition": ""
        },
        'nd_min': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'nd_max': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'vc_pow': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'vc_min': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'vc_max': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'lai_pow': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'diffusion_slope': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'diffusion_intercept': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        't_min': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        't_max': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'vp_slope': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'vp_offset': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'int_max': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'tenacity': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'rcan_max': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'ndvi_obs_min': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'ndvi_obs_max': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'obs_fr': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'z_obs': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'z_b': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'c1': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'iter_h': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'r_soil_pow': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'r_soil_min': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'se_top': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'porosity': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'r0_grass': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        'eps_a': {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "tau_co2_o2": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "year": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "t_air_k_min": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "t_air_k_max": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "t_air_k_12": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "t_dep": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "k_m": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "k_0": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "npp": {
            'long_name': '', 
            'units': 'gC m-2', 
            'definition': ''},
        "npp_max": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "co2_fert": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "f_par": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "apar": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "a_d": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "co2_act": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "t_dew_i": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
        "t_dew_24": {
            'long_name': '', 
            'units': '', 
            'definition': ''},
    }

    return defs


import numpy as np
from pywapor.et_look_dev import constants as c
import xarray as xr

def interception_wm2(int_mm, lh_24):
    r"""
    Computes the energy equivalent for the interception in Wm-2 if it
    is provide in mm/day

    .. math ::
        I = \frac{\lambda I^*}{86400}

    Parameters
    ----------
    int_mm : float
        interception
        :math:`I^*`
        [mm day-1]

    lh_24 : float
        daily latent heat for evaporation
        :math:`\lambda`
        [J kg-1]

    Returns
    -------
    int_wm2 : float
        interception
        :math:`I`
        [W m-2]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> import ETLook.meteo as meteo
    >>> lh = meteo.latent_heat_daily(20.0)
    >>> rad.interception_wm2(1.0, lh)
    28.40023148148148

    """
    return int_mm * (lh_24/c.day_sec)


def soil_fraction(lai):
    r"""
    Computes the effect of the vegetation has in separating the net radiation
    into a soil and canopy component. If the canopy has a full cover almost
    no radiation reaches the soil.

    .. math ::
        s_f = \exp\left(-0.6*I_{lai}\right)

    Parameters
    ----------
    lai : float
        leaf area index
        :math:`I_{lai}`
        [-]

    Returns
    -------
    sf_soil : float
        soil fraction
        :math:`s_f`
        [-]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.soil_fraction(3.0)
    0.16529888822158656
    """
    return np.exp(-0.6*lai)


def longwave_radiation_fao_etref(t_air_k_24, vp_24, trans_24):
    r"""
    Computes the net longwave radiation according to the FAO 56 manual. For the
    reference ET calculation the values for vp_slope, vp_offset, lw_slope and
    lw_offset are being provided as defaults

    .. math ::
        L^{*}=\sigma\left(T_{a,K}\right)^{4}
        \left(vp_off-vp_slp\sqrt{0.1e_{a}}
        \right)\left(lw_slp\frac{\tau}{0.75}+lw_off\right)

    where the following constant is used

    * :math:`\sigma` = Stefan Boltzmann constant = 5.67 e-8 J s-1 m-2 K-4

    Parameters
    ----------
    t_air_k_24 : float
        daily air temperature in Kelvin
        :math:`T_{a,K}`
        [-]
    vp_24 : float
        daily vapour pressure
        :math:`e_{a}`
        [mbar]
    trans_24 : float
        daily atmospheric transmissivity
        :math:`\tau`
        [-]

    Returns
    -------
    l_net : float
        daily net longwave radiation
        :math:`L^{*}`
        [Wm-2]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.longwave_radiation_fao_etref(t_air_k=302.5, vp=10.3, trans_24=0.6)
    68.594182173686306
    """

    vp_slope=0.14
    vp_offset=0.34
    lw_slope=1.35
    lw_offset=-0.35
    return longwave_radiation_fao(t_air_k_24, vp_24, trans_24, vp_slope, vp_offset, lw_slope, lw_offset)


def longwave_radiation_fao(t_air_k_24, vp_24, trans_24, vp_slope=0.14, vp_offset=0.34,
                           lw_slope=1.35, lw_offset=-0.35):
    r"""
    Computes the net longwave radiation according to the FAO 56 manual.

    .. math ::
        L^{*}=\sigma\left(T_{a,K}\right)^{4}
        \left(vp_{off}-vp_{slope}\sqrt{0.1e_{a}}
        \right)\left(lw_{slope}\frac{\tau}{0.75}+lw_{off}\right)

    where the following constant is used

    * :math:`\sigma` = Stefan Boltzmann constant = 5.67 e-8 J s-1 m-2 K-4

    Parameters
    ----------
    t_air_k_24 : float
        daily air temperature in Kelvin
        :math:`T_{a,K}`
        [-]
    vp_24 : float
        daily vapour pressure
        :math:`e_{a}`
        [mbar]
    trans_24 : float
        daily atmospheric transmissivity
        :math:`\tau`
        [-]
    vp_slope : float
        slope of the vp-term in the FAO-56 longwave radiation relationship
        :math:`vp_{slope}`
        [-]
    vp_offset : float
        offset of the vp-term in the FAO-56 longwave radiation relationship
        :math:`vp_{off}`
        [-]
    lw_slope : float
        slope of the tau-term in the FAO-56 longwave radiation relationship
        :math:`lw_{slope}`
        [-]
    lw_offset : float
        offset of the tau-term in the FAO-56 longwave radiation relationship
        :math:`lw_{off}`
        [-]

    Returns
    -------
    l_net : float
        daily net longwave radiation
        :math:`L^{*}`
        [Wm-2]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.longwave_radiation_fao(t_air_k=302.5, vp=10.3, trans_24=0.6)
    68.594182173686306
    """

    return c.sb*t_air_k_24**4*(vp_offset-vp_slope*np.sqrt(0.1*vp_24))*(lw_offset + lw_slope*(trans_24/0.75))


def net_radiation(r0, ra_24, l_net, int_wm2):
    r"""
    Computes the net radiation

    .. math ::
        Q^{*} = \left[\left(1-\alpha_{0}\right)S^{\downarrow}-L^{*}-I\right]

    Parameters
    ----------
    r0 : float
        albedo
        :math:`\alpha_{0}`
        [-]
    ra_24 : float
        daily solar radiation
        :math:`S^{\downarrow}`
        [Wm-2]
    l_net : float
        daily net longwave radiation
        :math:`L^{*}`
        [wm-2]
    int_wm2 : float
        interception
        :math:`I`
        [Wm-2]

    Returns
    -------
    rn_24 : float
        daily net radiation
        :math:`Q^{*}`
        [Wm-2]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.net_radiation(r0=0.10, ra_24=123., l_net=24., int_wm2=0)
    86.7
    """
    rn_24 = (1-r0)*ra_24-l_net-int_wm2
    rn_24 = rn_24.clip(20, 2000)
    return rn_24


def net_radiation_canopy(rn_24, sf_soil):
    r"""
    Computes the net radiation for the canopy

    .. math ::
        Q^{*}_{canopy} = \left(1-s_f\right) Q^{*}

    Parameters
    ----------
    rn_24 : float
        net radiation
        :math:`Q^{*}`
        [Wm-2]
    sf_soil : float
        soil fraction
        :math:`s_f`
        [-]

    Returns
    -------
    rn_24_canopy : float
        net radiation for the canopy
        :math:`Q^{*}_{canopy}`
        [Wm-2]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.net_radiation_canopy(rn_24=200, sf_soil=0.4)
    120.0

    """
    return rn_24 * (1-sf_soil)


def net_radiation_soil(rn_24, sf_soil):
    """
    Computes the net radiation for the soil

    .. math ::
        Q^{*}_{soil} = s_f Q^{*}

    Parameters
    ----------
    rn_24 : float
        net radiation
        :math:`Q^{*}`
        [Wm-2]
    sf_soil : float
        soil fraction
        :math:`s_f`
        [-]

    Returns
    -------
    rn_24_soil : float
        net radiation for the soil
        :math:`Q^{*}_{soil}`
        [Wm-2]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.net_radiation_soil(rn_24=200, sf_soil=0.4)
    80.0
    """
    return rn_24 * sf_soil


def net_radiation_grass(ra_24, l_net, r0_grass=0.23):
    r"""
    Computes the net radiation for reference grass

    .. math ::
        Q^{*} = \left[\left(1-\alpha_{0, grass}\right)S^{\downarrow}-L^{*}-I\right]

    Parameters
    ----------
    ra_24 : float
        daily solar radiation
        :math:`S^{\downarrow}`
        [Wm-2]
    l_net : float
        daily net longwave radiation
        :math:`L^{*}`
        [wm-2]
    r0_grass : float
        albedo for reference grass
        :math:`\alpha_{0, grass}`
        [-]

    Returns
    -------
    rn_24_grass : float
        daily net radiation for reference grass
        :math:`Q^{*}`
        [Wm-2]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.net_radiation_grass(ra_24=123., l_net=24.)
    70.7
    """
    return (1-r0_grass)*ra_24-l_net


def volumetric_heat_capacity(se_top=1.0, porosity=0.4):
    r"""
    Computes the volumetric heat capacity of the soil

    .. math ::
        \rho c_{p}=10e^{6}\left[\left(1-\phi\right)^{2}+
        2.5\phi+4.2\phi S_{e,top}\right]

    Parameters
    ----------
    se_top : float
        effective saturation of the topsoil
        :math:`S_{e,top}`
        [-]
    porosity : float
        porosity of the soil
        :math:`\phi`
        [-]

    Returns
    -------
    vhc : float
        volumetric heat capacity
        :math:`\rho c_{p}`
        [J m-3 K-1]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.volumetric_heat_capacity(se_top=0.4, porosity = 0.5)
    23400000.0
    """
    return ((1-porosity)**2+2.5*porosity+4.2*porosity*se_top)*10**6


def soil_thermal_conductivity(se_top):
    r"""
    Computes the soil thermal conductivity

    .. math ::
        k=0.15+18.5S_{e,top}

    Parameters
    ----------
    se_top : float
        effective saturation of the topsoil
        :math:`S_{e,top}`
        [-]

    Returns
    -------
    stc : float
        soil thermal conductivity
        :math:`k`
        [W m-1 K-1]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.soil_thermal_conductivity(se_top=0.4)
    0.8900000000000001
    """
    return 0.15 + 1.85 * se_top


def damping_depth(stc, vhc):
    r"""
    Computes the damping depth

    .. math ::
        z_{d}=\sqrt{\frac{2kP}{2\pi\rho c_{p}}}

    with the following constant

    * :math:`P` period (seconds within a year)

    Parameters
    ----------
    stc : float
        soil thermal conductivity
        :math:`k`
        [W m-1 K-1]
    vhc : float
        volumetric heat capacity
        :math:`\rho c_{p}`
        [J m-3 K-1]

    Returns
    -------
    dd : float
        damping depth
        :math:`z_{d}`
        [m]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.damping_depth(stc=0.9, vhc=volumetric_heat_capacity())
    0.54514600029013294
    """
    return np.sqrt((2*stc*c.year_sec)/(vhc*2*np.pi))


#TODO north-south transition with regard to latitude
def bare_soil_heat_flux(doy, dd, stc, t_amp_year, lat):
    r"""
    Computes the bare soil heat flux

    .. math ::
        G_{0}=\frac{\sqrt{2}A_{t,year}k\sin\left(\frac{2\pi J}{P}-
              \frac{\pi}{4}\right)}{z_{d}}

    where the following constant is used

    * :math:`P` period (seconds within a year)

    The term :math:`-\frac{\pi}{4}` is a phase shift for northern latitudes.
    For southern latitudes the phase shift will be :math:`-\frac{\pi}{4}+\pi`

    Parameters
    ----------
    stc : float
        soil thermal conductivity
        :math:`k`
        [W m-1 K-1]
    dd : float
        damping depth
        :math:`z_{d}`
        [m]
    t_amp_year : float
        yearly air temperature amplitude
        :math:`A_{t,year}`
        [m]
    doy : float
        julian day of the year
        :math:`J`
        [-]
    lat : float
        latitude
        :math:`\lambda`
        [rad]

    Returns
    -------
    g0_bs : float
        bare soil heat flux
        :math:`G_{0}`
        [m]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> stc = rad.soil_thermal_conductivity(se_top=1.0)
    >>> vhc = rad.volumetric_heat_capacity(se_top=1.0)
    >>> dd = damping_depth(stc,vhc)
    >>> rad.bare_soil_heat_flux(126, dd, stc, t_amp_year=13.4, lat=40*(math.pi/180.0))
    array([ 45.82350561])
    """

    if isinstance(lat, xr.DataArray):
        phase = xr.where(lat > 0, -np.pi/4.0, -np.pi/4.0+np.pi)
    else:
        phase = np.where(lat > 0, -np.pi/4.0, -np.pi/4.0+np.pi)

    out = (np.sqrt(2.0)*t_amp_year*stc*np.sin(2*np.pi/c.year_sec*doy*c.day_sec+phase))/dd

    return out


def soil_heat_flux(g0_bs, sf_soil, land_mask, rn_24_soil, trans_24, ra_24, l_net, rn_slope=0.92, rn_offset=-61.0):
    r"""
    Computes the soil heat flux

    .. math ::
        G=s_f G_{0}

    Parameters
    ----------
    g0_bs : float
        bare soil heat flux
        :math:`G_{0}`
        [W m-2]
    sf_soil : float
        soil fraction
        :math:`s_f`
        [-]
    land_mask : int
        land use classification
        :math:`l`
        [-]
    rn_24_soil : float
        net radiation for the soil
        :math:`Q^{*}_{soil}`
        [Wm-2]
    trans_24 : float
        daily atmospheric transmissivity
        :math:`\tau`
        [-]
    rn_slope : float
        slope rn/g0 relation water
        :math:`lws`
        [-]
    rn_offset : float
        offset rn/g0 relation water
        :math:`lwo`
        [-]
    ra_24 : float
        daily solar radiation
        :math:`S^{\downarrow}`
        [Wm-2]
    l_net : float
        daily net longwave radiation
        :math:`L^{*}`
        [wm-2]

    Returns
    -------
    g0_24 : float
        daily soil heat flux
        :math:`G`
        [W m-2]

    Examples
    --------
    >>> import ETLook.radiation as rad
    >>> rad.soil_heat_flux(g0_bs=12.4, sf_soil=0.4)
    4.960000000000001
    """
    def land_city_func(g0_bs, sf_soil):
        return g0_bs * sf_soil

    def water_func(ra_24, trans_24, l_net, rn_slope, rn_offset, rn_24_soil):
        rn_24_clear = 0.95 * ra_24 / trans_24 - l_net
        g0_24_clear = rn_24_clear * rn_slope + rn_offset
        g0_24_clear = np.minimum(g0_24_clear, 0.5 * rn_24_clear)

        # adjust water heat storage to current net radiation conditions
        g0_24 = g0_24_clear * rn_24_soil / rn_24_clear

        return g0_24

    if isinstance(land_mask, xr.DataArray):
        g0 = xr.zeros_like(land_mask)
        g0 = xr.where(land_mask == 1, land_city_func(g0_bs, sf_soil), g0)
        g0 = xr.where(land_mask == 2, water_func(ra_24, trans_24, l_net, rn_slope, rn_offset, rn_24_soil), g0)
        g0 = xr.where(land_mask == 3, land_city_func(g0_bs, sf_soil), g0)
    else:
        g0 = np.zeros_like(land_mask)
        g0 = np.where(land_mask == 1, land_city_func(g0_bs, sf_soil), g0)
        g0 = np.where(land_mask == 2, water_func(ra_24, trans_24, l_net, rn_slope, rn_offset, rn_24_soil), g0)
        g0 = np.where(land_mask == 3, land_city_func(g0_bs, sf_soil), g0)

    return g0



from pywapor.enhancers import gap_fill, apply_enhancers, dem, lulc, pressure, temperature, other, dms

__all__ = ['gap_fill', 'apply_enhancers', 'dem', 'lulc', 'pressure', 'temperature', 'other', 'dms']

__version__ = '0.1'


import os
from pywapor.collect.protocol import cog
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds
import xarray as xr
from functools import partial
from pywapor.enhancers import lulc
import numpy as np

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    variables = {
        '2009_V2.3_Global': {
                "Band1": [("lat", "lon"), "lulc"],
                "crs": [(), "spatial_ref"],
                    },
    }

    req_dl_vars = {
        "2009_V2.3_Global": {
            "lulc": ["Band1", "crs"],
            "rs_min": ["Band1", "crs"],
            "z_obst_max": ["Band1", "crs"], 
            "land_mask": ["Band1", "crs"],
            "lue_max": ["Band1", "crs"],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def remove_var(ds, var):
    return ds.drop_vars([var])

def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        '2009_V2.3_Global': {
            "lulc": [],
            "rs_min": [partial(lulc.lulc_to_x, in_var = "lulc", out_var = "rs_min", 
                        convertor = lulc.globcover_to_rs_min())],
            "z_obst_max": [partial(lulc.lulc_to_x, in_var = "lulc", out_var = "z_obst_max", 
                        convertor = lulc.globcover_to_z_obst_max())],
            "land_mask": [partial(lulc.lulc_to_x, in_var = "lulc", out_var = "land_mask", 
                        convertor = lulc.globcover_to_land_mask())],
            "lue_max": [partial(lulc.lulc_to_x, in_var = "lulc", out_var = "lue_max", 
                        convertor = lulc.globcover_to_lue_max())],
            },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def url_func(product_name):
    """Returns a url at which to collect GLOBCOVER data.

    Parameters
    ----------
    product_name : None
        Not used.

    Returns
    -------
    str
        The url.
    """
    return r"http://due.esrin.esa.int/files/GLOBCOVER_L4_200901_200912_V2.3.color.tif"

def download(folder, latlim, lonlim, product_name, req_vars = ["lulc"],
                variables = None, post_processors = None, **kwargs):
    """Download GLOBCOVER data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str, optional
        Name of the product to download, by default "P05".
    req_vars : list, optional
        Which variables to download for the selected product, by default ["p"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """

    folder = os.path.join(folder, "GLOBCOVER")

    appending = False
    fn = os.path.join(folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]

    coords = {"x": ("lon", lonlim), "y": ("lat", latlim)}

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    ds_new = cog.download(fn, product_name, coords, variables, 
                        post_processors, url_func, ndv = 0)

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds

if __name__ == "__main__":

    product_name = '2009_V2.3_Global'

    folder = r"/Users/hmcoerver/Local/globcov_test"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    variables = None
    post_processors = None
    req_vars = ["lulc", "rs_min", "land_mask"]

    ds = download(folder, latlim, lonlim, product_name, req_vars = req_vars,
                    variables = variables, post_processors = post_processors)



"""
    The leaf module contains all functions related to estimating vegetation
    cover. These functions only work on an instantaneous basis.

"""
import numpy as np
import xarray as xr

def vegetation_cover(ndvi, nd_min=0.125, nd_max=0.8, vc_pow=0.7):
    r"""
    Computes vegetation cover based on NDVI

    .. math ::
        c_{veg} =
        \begin{cases}
        \begin{array}{cc}
        0 & I_{NDVI}\leq I_{NDVI,min}\\
        1-\left(\frac{I_{NDVI,max}-I_{NDVI}}{I_{NDVI,max}-I_{NDVI,min}}\right)^
        {a} & I_{NDVI,min}<I_{NDVI}<I_{NDVI,max}\\
        1 & I_{NDVI}\geq I_{NDVI,max}
        \end{array}\end{cases}

    Parameters
    ----------
    ndvi : float
        Normalized Difference Vegetation Index
        :math:`I_{NDVI}`
        [-]
    nd_min : float
        NDVI value where vegetation cover is 0
        :math:`I_{NDVI,min}`
        [-]
    nd_max : float
        NDVI value where vegetation cover is 1
        :math:`I_{NDVI,max}`
        [-]
    vc_pow : float
        Exponential power used in vegetation cover function
        :math:`a`
        [-]

    Returns
    -------
    vc : float
        vegetation cover
        :math:`c_{veg}`
        [-]

    Examples
    --------
    >>> from ETLook import leaf
    >>> leaf.vegetation_cover(0.1, nd_min=0.2)
    0
    >>> leaf.vegetation_cover(0.5)
    0.4331446663885373
    >>> leaf.vegetation_cover(0.85)
    1

    """

    if np.isscalar(ndvi):

        if ndvi <= nd_min:
            res = 0
        if (ndvi > nd_min) & (ndvi < nd_max):
            res = 1 - ((nd_max - ndvi) / (nd_max - nd_min)) ** vc_pow
        if ndvi >= nd_max:
            res = 1

    else:

        if isinstance(ndvi, np.ndarray):
            # Create empty array
            res = np.ones(ndvi.shape) * np.nan

            # fill in array
            res[ndvi <= nd_min] = 0
            res[np.logical_and(ndvi > nd_min, ndvi < nd_max)] = 1 - (
                        (nd_max - ndvi[np.logical_and(ndvi > nd_min, ndvi < nd_max)]) / (nd_max - nd_min)) ** vc_pow
            res[ndvi >= nd_max] = 1
        else:
            res = 1 - ((nd_max - ndvi) / (nd_max - nd_min)) ** vc_pow
            res = res.clip(0.0, 1.0)

    return res


def leaf_area_index(vc, vc_min=0.0, vc_max=vegetation_cover(0.795), lai_pow=-0.45):
    r"""
    Computes leaf area index based on vegetation cover. It is based on the
    Kustas formulation of LAI vs NDVI.

    .. math ::
        I_{lai} =
        \begin{cases}
        \begin{array}{cc}
        0 & c_{veg}\leq c_{veg,min}\\
        \frac{\ln\left(-\left(c_{veg}-1\right)\right)}{b} &
        c_{veg,min}<c_{veg}\leq c_{veg,max}\\
        \frac{\ln\left(-\left(c_{veg,max}-1\right)\right)}{b} &
        c_{veg}>c_{veg,max}
        \end{array}\end{cases}

    Parameters
    ----------
    vc : float
        vegetation cover
        :math:`c_{veg}`
        [-]
    vc_min : float
        vegetation cover where LAI is 0
        :math:`c_{veg,min}`
        [-]
    vc_max : float
        vegetation cover at maximum LAI
        :math:`c_{veg,max}`
        [-]
    lai_pow : float
        exponential factor used in LAI function
        :math:`b`
        [-]

    Returns
    -------
    lai : float
        leaf area index
        :math:`I_{lai}`
        [-]

    Examples
    --------
    >>> from ETLook import leaf
    >>> leaf.leaf_area_index(0.0)
    0
    >>> leaf.leaf_area_index(0.5)
    1.5403270679109895
    >>> leaf.leaf_area_index(1.0)
    7.6304274331264414

    """

    if np.isscalar(vc):
        if vc <= vc_min:
            res = 0
        if (vc > vc_min) & (vc < vc_max):
            res = np.log(-(vc - 1)) / lai_pow
        if vc >= vc_max:
            res = np.log(-(vc_max - 1)) / lai_pow

    elif isinstance(vc, xr.DataArray):
        res = np.ones(vc.shape) * np.nan
        res = xr.where(vc <= vc_min, 0, res)
        res = xr.where((vc > vc_min) & (vc < vc_max), 
                        np.log(-(vc - 1)) / lai_pow, res)
        res = xr.where(vc >= vc_max, np.log(-(vc_max - 1)) / lai_pow, res)
    else:
        # Create empty array
        res = np.ones(vc.shape) * np.nan

        # fill in array
        res[vc <= vc_min] = 0
        res[np.logical_and(vc > vc_min, vc < vc_max)] = np.log(
            -(vc[np.logical_and(vc > vc_min, vc < vc_max)] - 1)) / lai_pow
        res[vc >= vc_max] = np.log(-(vc_max - 1)) / lai_pow

    return res


def effective_leaf_area_index(lai):
    r"""
    Computes effective leaf area index, this describes the leaf area which
    actively participates in transpiration. It is based on the actual leaf
    area index and an extinction function. So with a higher leaf area index the
    effective leaf area index is a smaller percentage of the total leaf area
    index.

    .. math ::
        I_{lai,eff}=\frac{I_{lai}}{0.3I_{lai}+1.2}

    Parameters
    ----------
    lai : float
        Leaf area index
        :math:`I_{lai}`
        [-]

    Returns
    -------
    lai_eff : float
        effective leaf area index
        :math:`I_{lai,eff}`
        [-]

    Examples
    --------
    >>> from ETLook import leaf
    >>> leaf.effective_leaf_area_index(3.0)
    1.4285714285714288
    >>> leaf.effective_leaf_area_index(5.0)
    1.8518518518518516

    """
    lai_eff = lai / ((0.3 * lai) + 1.2)
    return lai_eff

def fpar(vc, ndvi):
    r"""
    Computes the fpar

    Parameters
    ----------
    vc : float
        vegetation cover
        :math:`c_{veg}`
        [-]
    ndvi : float
        Normalized Difference Vegetation Index
        :math:`I_{NDVI}`
        [-]
        
    Returns
    -------
    fpar : float
        fpar
        :math:`I_{vc,ndvi}`
        [-]

    """
    # fpar = 0.925 * vc
    fpar = 0.8624 * ndvi - 0.0814
    if isinstance(fpar, xr.DataArray):
        fpar = xr.where(ndvi < 0.1, 0.0, fpar)
    else:
        fpar[ndvi < 0.1] = 0.0

    return fpar

def apar(ra_24, fpar):    
    r"""
    Computes the Aborbed Photosynthetical Active Radiation

    Parameters
    ----------
    ra_24 : float
        daily solar radiation
        :math:`S^{\downarrow}`
        [Wm-2]
    fpar : float
        fpar
        :math:`I_{vc,ndvi}`
        [-]
        
    Returns
    -------
    apar : float
        apar
        :math:`I_{ra_24,fpar}`
        [MJ/m2]

    """
    par = 0.48 * ra_24 * 0.0864
    apar = par * fpar
    
    return apar     
    


from pywapor.general.logger import log
from pywapor.collect.protocol.projections import get_crss
import pywapor.collect.protocol.opendap as opendap
import pywapor.collect.accounts as accounts
import os
import xarray as xr
import numpy as np
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds

def default_vars(product_name, req_vars = ["p"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["p"].

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    
    variables =  {
        "P05": {
            "precip": [("time", "latitude", "longitude"), "p"],
        }
    }

    req_dl_vars = {
        "P05": {
            "p": ["precip"],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def default_post_processors(product_name, req_vars = ["p"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["p"].

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "P05": {
            "p": [],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def fn_func(product_name, tile):
    """Returns a filename for the `product_name`.

    Parameters
    ----------
    product_name : str
        Name of the product.
    tile : None
        Not used.

    Returns
    -------
    str
        Filename.
    """
    fn = f"{product_name}_temp.nc"
    return fn

def url_func(product_name, tile):
    """Returns a url at which to collect CHIRPS data.

    Parameters
    ----------
    product_name : None
        Not used.
    tile : None
        Not used.

    Returns
    -------
    str
        The url.
    """
    url = "https://coastwatch.pfeg.noaa.gov/erddap/griddap/chirps20GlobalDailyP05.nc?"
    return url

def download(folder, latlim, lonlim, timelim, product_name = "P05", req_vars = ["p"],
                variables = None, post_processors = None):
    """Download CHIRPS data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str, optional
        Name of the product to download, by default "P05".
    req_vars : list, optional
        Which variables to download for the selected product, by default ["p"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "CHIRPS")

    appending = False
    fn = os.path.join(folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]

    tiles = [None]
    coords = {"x": ["longitude", lonlim], "y": ["latitude", latlim], "t": ["time", timelim]}
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    timedelta = np.timedelta64(30, "m")
    data_source_crs = get_crss("WGS84")
    parallel = False
    spatial_tiles = False
    un_pw = accounts.get("NASA")
    request_dims = False
    ds_new = opendap.download(fn, product_name, coords, 
                variables, post_processors, fn_func, url_func, un_pw = un_pw, 
                tiles = tiles, data_source_crs = data_source_crs, parallel = parallel, 
                spatial_tiles = spatial_tiles, request_dims = request_dims,
                timedelta = timedelta)


    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds

if __name__ == "__main__":

    import datetime

    folder = r"/Users/hmcoerver/Downloads/pywapor_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 11)]

    product_name = "P05"
    req_vars = ["p"]
    # CHIRPS.
    ds = download(folder, latlim, lonlim, timelim)

